#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import sys

def xy2rp(x, y):
    """Convert (x,y) to (R,phi) in polar coordinate

    Args:
        x (float): x coordinate
        y (float): y coordinate

    Returns:
        R (float): radius
        phi (float): angle in rad
    """    
    R = np.sqrt(x**2 + y**2)
    if   x >  0.0 and y >= 0.0 : # [0,pi/2)
        phi = np.arcsin(y/R)
    elif x <= 0.0 and y >  0.0 : # [pi/2, pi)
        phi = np.arccos(x/R)
    elif x <  0.0 and y <= 0.0 : # [pi, 3/2 pi)
        phi = np.arccos(-x/R) + np.pi
    elif x >= 0.0 and y <  0.0 : # [3/2 pi, 2pi)
        phi = np.arcsin(y/R) + 2*np.pi
    else:
        raise ValueError("Something wrong with your inputs ({:f}, {:f}).".format(x, y))
    return R, phi

def map_matrix(xx, first=True, second=True):
    """Map matrix to be complete (closed)
    Arguments:
      xx -- 2D numpy array
      first -- boolean, default: True, if increase the first dimension
      second -- boolean, default: True, if increase the second dimension

    Returns:
      new -- the new matrix with dimension increased
    """
    a, b = np.shape(xx)
    # only first
    if first and not second :
        new = np.zeros((a+1,b))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[  a, 0:b] = xx[0  , 0:b]
    # only second
    elif not first and second :
        new = np.zeros((a,b+1))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[0:a,   b] = xx[0:a, 0  ]
    # both direction
    elif first and second :
        new = np.zeros((a+1,b+1))
        new[0:a, 0:b] = xx[0:a, 0:b]
        new[  a, 0:b] = xx[0  , 0:b]
        new[0:a,   b] = xx[0:a, 0  ]
        new[  a,   b] = xx[0  , 0  ]
    # otherwise return the original matrix
    else :
        return xx
    return new

def toroidal_period(vec, nfp=1):
    """
    vec: [x,y,z] data
    Nfp: =1, toroidal number of periodicity
    """
    phi = 2*np.pi/nfp
    vec = np.atleast_2d(vec)
    new_vec = vec.copy()
    for ifp in range(nfp):
        if ifp==0:
            continue
        rotate = np.array([[  np.cos(ifp*phi), np.sin(ifp*phi), 0], \
                           [ -np.sin(ifp*phi), np.cos(ifp*phi), 0], \
                           [                0,               0, 1]])
        new_vec = np.concatenate((new_vec, np.matmul(vec, rotate)))
    return new_vec

def print_progress(iteration, total, prefix='Progress', suffix='Complete', decimals=1, bar_length=60):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """
    str_format = "{0:." + str(decimals) + "f}"
    percents = str_format.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '█' * filled_length + '-' * (bar_length - filled_length)

    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),

    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()
    return

# Smart way to check where to plot
def get_figure(axes=None, **kwargs):
    """
    Check where to plot
    Parameters:
        axes: matplotlib.pyplot axis, axis to plot (default: None)
        kwargs: keyword arguments
    Return:
        f, ax 
        f  : matplotlib.pyplot figure
        ax : matplotlib.pyplot axis 
    """
    import matplotlib.pyplot as plt
    if axes is None:
        # No axes provided
        f, axes = plt.subplots(**kwargs)
        '''
        f = plt.gcf()
        if len(f.axes):
            # normal situation in which existing figures should be respected and left alone
            f, axes = plt.subplots(**kwargs)
        else:
            #  made a empty figure for using
            axes = f.add_subplot(**kwargs)
         '''
    else:
        # axes = np.atleast_1d(axes)
        f = axes.get_figure()
    return f, axes

def colorbar(mappable, **kwargs):
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    import matplotlib.pyplot as plt
    last_axes = plt.gca()
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(mappable, cax=cax, **kwargs)
    plt.sca(last_axes)
    return cbar

def kwargs2dict(**kwargs):
    return kwargs

def vmecMN(mpol, ntor):
# manipulate VMEC index
    mn = (2*ntor+1)*mpol - ntor # total number of Fourier harmonics
    xm = np.zeros((mn,), dtype=int)
    xn = np.zeros((mn,), dtype=int)
    imn = 0 
    for ii in range(mpol):
        for jj in range(-ntor, ntor+1):
            if ii==0 and jj<0 :
                continue
            xm[imn] = ii
            xn[imn] = jj
            imn += 1
    return xm, xn

def trigfft(y, tr=-1):
    """calculate trigonometric coefficients using FFT
    Assuming the periodicity is 2*pi
    params:
        y -- 1D array for Fourier transformation
        tr -- Truncation number (default: -1)
    return:
        a dict containing
        'n' -- index
        'rcos' -- cos coefficients of the real part
        'rsin' -- sin coefficients of the real part
        'icos' -- cos coefficients of the imag part
        'isin' -- sin coefficients of the imag part
    """
    from scipy.fftpack import fft
    N = len(y)
    if N%2 == 0 : # even
        half = N//2-1
        end = half + 2
    else :
        half = (N-1)//2
        end = half + 1
    assert tr <= end, 'Truncation number should be smaller than dimension!'
    comp = fft(y)/N
    a_k = np.zeros(end, dtype=np.complex)
    b_k = np.zeros(end, dtype=np.complex)
    a_k[0] = comp[0]
    for n in range(1, half+1):
        a_k[n] =  comp[n] + comp[N-n]
        b_k[n] = (comp[n] - comp[N-n]) * 1j
    if N%2 == 0: # even number
        a_k[end-1] = comp[N//2]
    index = np.arange(end)
    
    return {'n': index[:tr],
            'rcos': np.real(a_k[:tr]),
            'rsin': np.real(b_k[:tr]), 
            'icos': np.imag(a_k[:tr]), 
            'isin': np.imag(b_k[:tr])
           }

def fft_deriv(y):
    from scipy.fftpack import fft, ifft
    N = len(y)
    comp = fft(y)
    if N%2 == 0:
        dt = (np.arange(N) - np.concatenate((np.zeros(N//2), [N//2], N*np.ones(N//2-1))))*1j
    else:
        dt = (np.arange(N) - np.concatenate((np.zeros(N//2), N*np.ones(N//2+1))))*1j
    return ifft(comp*dt)

def trig2real(theta, zeta, xm, xn, fmnc=None, fmns=None):
        npol, ntor = len(theta), len(zeta)
        _tv, _zv = np.meshgrid(theta, zeta, indexing='ij')
        # mt - nz (in matrix)
        _mtnz = np.matmul( np.reshape(xm, (-1,1)), np.reshape(_tv, (1,-1)) ) \
              - np.matmul( np.reshape(xn, (-1,1)), np.reshape(_zv, (1,-1)) ) 
        _cos = np.cos(_mtnz)
        _sin = np.sin(_mtnz)

        f = np.zeros((1, npol*ntor))
        if fmnc is not None:
            f += np.matmul(np.reshape(fmnc, (1,-1)), _cos)
        if fmns is not None:
            f += np.matmul(np.reshape(fmns, (1,-1)), _sin)
        return f.reshape(npol, ntor)     



import h5py
import numpy as np # for isscalar
import os          # for path.abspath
import keyword     # for getting python keywords

# Modified from J. Schilling (jonathan.schilling@ipp.mpg.de)'s script for SPEC

class HDF5:  
    # use as s = HDF5(filename), e.g. s=HDF5("ext.h5") 
    def __init__(self, *args, **kwargs):
        # args[0] should always be the name of a file or an item inside the root object
        # if args[0] is not a filename, kwargs['content'] should be the content to be added
        # as self.`args[0]`
        
        _content = None
        if kwargs.get('content') == None:
            # assume arg[0] is a filename
            _content = h5py.File(args[0], "r")
            
            # keep track of which file this object corresponds to
            self.filename = os.path.abspath(args[0])
        elif isinstance(kwargs['content'], h5py.Group):
            _content = kwargs['content']
        
        if (_content != None):
            for key in _content:
                if isinstance(_content[key], h5py.Group):
                    # recurse into group
                    setattr(self, key, HDF5(content=_content[key]))
                elif isinstance(_content[key], h5py.Dataset): # read dataset
                    if key in keyword.kwlist: # avoid assign python keywords
                        setattr(self, key+'1', _content[key][()])
                    else: # if just one element, use the value directly
                        if len(_content[key][()]) == 1: 
                            setattr(self, key, _content[key][0])
                        else: 
                            setattr(self, key, _content[key][()])
                    # glue string together
                    if isinstance(_content[key][0], bytes): 
                        abc = ''
                        for i in _content[key]:
                            abc += i.decode('utf-8')
                        setattr(self, key, abc)
        
        if isinstance(_content, h5py.File):
            _content.close()
            
    # needed for iterating over the contents of the file
    def __iter__(self):
        return iter(self.__dict__)
    def __next__(self):
        return next(self.__dict__)
    
    # print a list of items contained in this object
    def inventory(self, prefix=""):
        _prefix = ""
        if prefix != "":
            _prefix = prefix+"/"
        
        for a in self:
            try:
                # recurse into member
                getattr(self, a).inventory(prefix=_prefix+a)
            except:
                # print item name
                print(_prefix+a)
    # delete
    def __del__(self):
        class_name = self.__class__.__name__



from .hdf5 import HDF5
from .misc import get_figure, map_matrix
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import FuncFormatter

class FOCUSHDF5(HDF5):
    """FOCUS output hdf5 file

    """
    # initialization, test = FOCUSHDF5('focus_test.h5')
    def __init__(self, filename, periodic=False, **kwargs):
        """ Initialization

        Keyword arguments:
            filenmae -- string, path and name to FOCUS output hdf5 file,
                        usually in the format of 'focus_*.h5'
            periodic -- logical, map all 2D surface data automatically, 
                        (default: True)
        """
        # read data
        super().__init__(filename)
        # print version
        try:
            print("FOCUS version: "+self.version)
        except AttributeError:
            print(filename + " is not a valid FOCUS output. Please check.")
            raise
        # add additional colume and row for plotting
        if periodic:
            self.xsurf = map_matrix(self.xsurf)
            self.ysurf = map_matrix(self.ysurf)
            self.zsurf = map_matrix(self.zsurf)
            self.nx = map_matrix(self.nx)
            self.ny = map_matrix(self.ny)
            self.nz = map_matrix(self.nz)
            self.nn = map_matrix(self.nn)
            self.Bx = map_matrix(self.Bx)
            self.By = map_matrix(self.By)
            self.Bz = map_matrix(self.Bz)
            self.Bn = map_matrix(self.Bn)
            self.plas_Bn = map_matrix(self.plas_Bn)
            
    # convergence plot
    def convergence(self, term='bnorm', iteration=True, axes=None, **kwargs):
        # get figure
        fig, axes = get_figure(axes)
        # set default plotting parameters
        kwargs['linewidth'] = kwargs.get('linewidth', 2.5) # line width
        #kwargs['marker'] = kwargs.get('marker', 'o') # extent
        # get iteration data
        if iteration:
            abscissa = 1+ np.arange(self.iout)
            _xlabel = 'iteration'
        else :
            abscissa = self.evolution[0,:] # be careful; DF is not saving wall-time
            _xlabel = 'wall time [Second]'
        # plot data
        if term.lower() == 'chi':
            data = self.evolution[1,:]
            kwargs['label'] = kwargs.get('label', r'$\chi^2$') # line label
        elif term.lower() == 'gradient':
            data = self.evolution[2,:]
            kwargs['label'] = kwargs.get('label', r'$|d \chi^2 / d {\bf X}|$') # line label
        elif term.lower() == 'bnorm':
            data = self.evolution[3,:]
            kwargs['label'] = kwargs.get('label', r'$f_{B_n}$') # line label
        elif term.lower() == 'bharm':
            data = self.evolution[4,:]
            kwargs['label'] = kwargs.get('label', r'$f_{B_{mn}}$') # line label
        elif term.lower() == 'tflux':
            data = self.evolution[5,:]
            kwargs['label'] = kwargs.get('label', r'$f_{\Psi}$') # line label
        elif term.lower() == 'ttlen':
            data = self.evolution[6,:]
            kwargs['label'] = kwargs.get('label', r'$f_L$') # line label
        elif term.lower() == 'cssep':
            data = self.evolution[7,:]
            kwargs['label'] = kwargs.get('label', r'$f_{CS}$') # line label
        elif term.lower() == 'curv':
            data = self.evolution[8,:]
            kwargs['label'] = kwargs.get('label', r'$f_{curv}$') # line label
        elif term.lower() == 'all':
            lines = []
            lines.append(self.convergence(term='chi', iteration=iteration, axes=axes, 
                                    linestyle='-', color='k', **kwargs))
            lines.append(self.convergence(term='gradient', iteration=iteration, axes=axes, 
                                    linestyle='--', color='k', **kwargs))
            lines.append(self.convergence(term='bnorm', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='r', **kwargs))
            lines.append(self.convergence(term='tflux', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='g', **kwargs))
            lines.append(self.convergence(term='bharm', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='b', **kwargs))
            lines.append(self.convergence(term='ttlen', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='c', **kwargs))
            lines.append(self.convergence(term='cssep', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='m', **kwargs))    
            lines.append(self.convergence(term='curv', iteration=iteration, axes=axes, 
                                    linestyle='-.', color='y', **kwargs))  
            #fig.legend(loc='upper right', frameon=False, ncol=2, prop={'size':16})
            plt.legend()
            return lines
        else :
            raise ValueError('unsupported option for term') 
        line = axes.semilogy(abscissa, data, **kwargs)
        axes.tick_params(axis='both', which='major', labelsize=15)
        axes.set_xlabel(_xlabel, fontsize=15)
        axes.set_ylabel('cost functions', fontsize=15)
        # fig.legend(loc='upper right', frameon=False, prop={'size':24, 'weight':'bold'})
        plt.legend()
        return line
    
    # poincare plot
    def poincare_plot(self, color=None, **kwargs):
        """Poincare plot
        Args:
             color (matplotlib color, or None): dot colors; default None (rainbow).
             kwargs : matplotlib scatter keyword arguments
             
        Returns:
             None
        """
        import matplotlib.pyplot as plt
        from matplotlib import cm
        
        # get figure and ax data
        if plt.get_fignums():
            fig = plt.gcf()
            ax = plt.gca()
        else :
            fig, ax = plt.subplots()
            
        # get colors
        if color == None:
            colors = cm.rainbow(np.linspace(1, 0, self.pp_ns))
        else :
            colors = [color]*self.pp_ns
        kwargs['s'] = kwargs.get('s',0.1) # dotsize
        # scatter plot
        for i in range(self.pp_ns):
            ax.scatter(self.ppr[:,i], self.ppz[:,i], color=colors[i], **kwargs)
        plt.axis('equal')
        plt.xlabel('R [m]',fontsize=20)
        plt.ylabel('Z [m]',fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        return 
    
    # Bnorm plot
    def Bnorm(self, plottype='2D', source='all', axes=None, flip=False, **kwargs):
        """ Plot Bn distribution.

        Keyword arguments:
            plottype -- string, '2D' (default) or '3D', determine the plottype
            source -- string, 'coil', 'plasma', 'sum' or 'all' (default), data source
            axes  -- matplotlib.pyplot or mayavi.mlab axis, axis to be 
                     plotted on  (default None)
            flip -- logical, determine how to calculate Bn from coil,
                    True: coil_Bn = Bn - plas_Bn; False: coil_Bn = Bn + plas_Bn

        Returns:
           obj -- matplotlib.pyplot or mayavi.mlab plotting object   
        """
        obj = []
        # check coil_Bn
        if flip:
            coil_Bn = self.Bn - self.plas_Bn
        else :
            coil_Bn = self.Bn + self.plas_Bn
        # 2D plots
        if plottype.lower() == '2d':
            # prepare coordinates
            ntheta  = self.Nteta
            nzeta = self.Nzeta
            nt, nz = self.Bn.shape
            if self.IsSymmetric == 0:
                zeta_end = 2*np.pi
            elif self.IsSymmetric == 1:
                zeta_end = 2*np.pi/self.Nfp
            elif self.IsSymmetric == 2:
                zeta_end = np.pi/self.Nfp
            else:
                raise ValueError('Something wrong with IsSymmetric.')
            def theta(x, pos):
                return '{:3.2f}'.format(np.pi/ntheta + x*(2*np.pi)/ntheta)
            def zeta(x, pos):
                return '{:3.2f}'.format(0.5*zeta_end/nzeta + x*zeta_end/nzeta)
            theta_format = FuncFormatter(theta)
            zeta_format = FuncFormatter(zeta)
            # planar plotting
            if source.lower() == 'all':
                fig, axes = get_figure(axes, ncols=3, sharex=True)
                axes = np.atleast_1d(axes)
                #kwargs['aspect'] = kwargs.get('aspect', nt/(2.0*nz*self.Nfp))
                kwargs['aspect'] = kwargs.get('aspect', 'auto')
            else :
                fig, axes = get_figure(axes)
                axes = np.atleast_1d(axes)
                kwargs['aspect'] = kwargs.get('aspect', float(nt/nz))
            # prepare axes
            for ax in axes:
                ax.xaxis.set_major_formatter(zeta_format)
                ax.yaxis.set_major_formatter(theta_format)
            # set default plotting parameters
            kwargs['cmap'] = kwargs.get('cmap','RdBu_r') # colormap
            kwargs['origin'] = kwargs.get('origin', 'lower') # number of contours
            kwargs['extent'] = kwargs.get('extent', [0, nt, 0, nz]) # extent
            plt.subplots_adjust(hspace=0.05)
            # imshow 
            if source.lower() == 'coil':
                obj.append(axes[0].imshow(np.transpose(coil_Bn), **kwargs))
                axes[0].set_title('Bn from coils', fontsize=15)
                axes[0].set_ylabel(r'$\theta$', fontsize=14)
                axes[0].set_xlabel(r'$\phi$', fontsize=14)
            elif source.lower() == 'plasma':
                obj.append(axes[0].imshow(np.transpose(self.plas_Bn), **kwargs))
                axes[0].set_title('Bn from plasma', fontsize=15)
                axes[0].set_ylabel(r'$\theta$', fontsize=14)
                axes[0].set_xlabel(r'$\phi$', fontsize=14)
            elif source.lower() == 'sum':
                obj.append(axes[0].imshow(np.transpose(self.Bn), **kwargs))
                axes[0].set_title('Residual Bn', fontsize=15)
                axes[0].set_ylabel(r'$\theta$', fontsize=14)
                axes[0].set_xlabel(r'$\phi$', fontsize=14)
            elif source.lower() == 'all':
                vmin = np.min([self.plas_Bn, self.Bn, coil_Bn])
                vmax = np.max([self.plas_Bn, self.Bn, coil_Bn])
                obj.append(axes[0].imshow(np.transpose(self.plas_Bn), 
                                          vmin=vmin, vmax=vmax, **kwargs))
                #plt.setp(axes[0].get_xticklabels(), visible=False)
                axes[1].set_title('Bn from plasma (left), coil (mid), overall(right).', fontsize=14) 
                obj.append(axes[1].imshow(np.transpose(coil_Bn),
                                          vmin=vmin, vmax=vmax, **kwargs))
                #axes[1].set_title('Bn from coils', fontsize=15)  
                plt.setp(axes[1].get_yticklabels(), visible=False)
                obj.append(axes[2].imshow(np.transpose(self.Bn), 
                                          vmin=vmin, vmax=vmax, **kwargs))
                plt.setp(axes[2].get_yticklabels(), visible=False)
                #axes[2].set_title('Residual Bn', fontsize=15)
                axes[0].set_ylabel(r'$\theta$', fontsize=14)
                #axes[1].set_ylabel(r'$\theta$', fontsize=14)
                #axes[2].set_ylabel(r'$\theta$', fontsize=14)
                axes[0].set_xlabel(r'$\phi$', fontsize=14)
                axes[1].set_xlabel(r'$\phi$', fontsize=14)
                axes[2].set_xlabel(r'$\phi$', fontsize=14)
            else:
                raise ValueError('unsupported option for source')       
            fig.subplots_adjust(right=0.85)
            cbar_ax = fig.add_axes([0.86, 0.10, 0.04, 0.8])
            fig.colorbar(obj[0], cax=cbar_ax)    
        return obj
    
    # Bmod plot
    def Bmod(self):
        return
    
    # write vtk
    def toVTK(self, name=None, full=False, **kwargs):
        """Save surface and magnetic field data into VTK file
        Arguments:
          name -- string, VTK file name. default: None, if None, using self.filename
          full -- boolean, if save the entire torus
          **kwargs -- external data will be saved in VTK file

        Return:
          VTK file name
        """
        from pyevtk.hl import gridToVTK
        # automatically get a file name, focus_*.h5 -> vtk_*
        if name is None:
            name = self.filename[:-3].replace('focus_', 'vtk_')
        xx = self.xsurf ; yy = self.ysurf ; zz = self.zsurf
        Bx = self.Bx    ; By = self.By    ; Bz = self.Bz
        Bn = self.Bn
        plas_Bn = self.plas_Bn
        toroidal = full
        xx = np.atleast_3d(map_matrix(xx, first=toroidal))
        yy = np.atleast_3d(map_matrix(yy, first=toroidal))
        zz = np.atleast_3d(map_matrix(zz, first=toroidal))
        Bn = np.atleast_3d(map_matrix(Bn, first=toroidal))
        plas_Bn =  np.atleast_3d(map_matrix(plas_Bn, first=toroidal))
        B = (np.atleast_3d(map_matrix(Bx, first=toroidal)),
             np.atleast_3d(map_matrix(By, first=toroidal)),
             np.atleast_3d(map_matrix(Bz, first=toroidal)))
        if full :
            pass
        data = {"Bn":Bn, "plas_Bn":plas_Bn, "B":B}
        data.update(kwargs)
        return gridToVTK(name, xx, yy, zz, pointData=data)
    
    



"""
This is a cutomized python package for plotting and data processing in stellarator optimization.

how to use:

"""
__version__ = '0.2.17'

# local packages
from .misc import colorbar, get_figure, kwargs2dict, map_matrix
from .misc import print_progress, toroidal_period, vmecMN, xy2rp
from .misc import trigfft, fft_deriv, trig2real
from .hdf5 import HDF5
from .surface import FourSurf
from .dipole import Dipole
from .focushdf5 import FOCUSHDF5
from .coils import Coil, SingleCoil
from .stellopt import STELLout
from .vmec import VMECout
from .booz_xform import BOOZ_XFORM



# modified from stellopt.pySTEL

from __future__ import print_function, absolute_import
from builtins import map, filter, range
from .sortedDict import SortedDict
import numpy as np 
import matplotlib.pyplot as plt
from .misc import get_figure

__all__ = ['STELLout', 'OMFITascii']

class OMFITascii(object):
    """
    OMFIT class used to interface with ASCII files

    :param filename: filename passed to OMFITobject class

    :param fromString: string that is written to file

    :param \**kw: keyword dictionary passed to OMFITobject class
    """
    def __init__(self, filename, **kw):
        fromString = kw.pop('fromString', None)
        self.filename = filename
        if fromString is not None:
            with open(self.filename, 'wb') as f:
                f.write(fromString.encode('utf-8'))

    def read(self):
        '''
        Read ASCII file and return content

        :return: string
        '''
        return open(self.filename, 'r').read()

    def write(self, value):
        '''
        Write string value to ASCII file

        :param value: string to be written to file

        :return: string
        '''
        open(self.filename, 'w').write(value)
        return value

    def append(self, value):
        '''
        Append string value to ASCII file

        :param value: string to be written to file

        :return: string
        '''
        open(self.filename, 'a').write(value)
        return self.read()

class STELLout(SortedDict, OMFITascii):
    """
    OMFITobject used to interface with stellopt.* file in STELLOPT outputs.

    :param filename: filename passed to OMFITascii class

    All additional key word arguments passed to OMFITascii

    """

    def __init__(self, filename, **kwargs):
        SortedDict.__init__(self)
        OMFITascii.__init__(self, filename, **kwargs)
        self.dynaLoad = True

    def load(self):
        """Load the file and parse it into a sorted dictionary"""
        file_handle = open(self.filename,'r')
        niter = 0
        for line in file_handle:
            if 'ITER' in line:
                niter=niter+1
            if 'MIN' in line:
                niter=niter-1
        self['ITER'] = np.ndarray((niter,1));
        file_handle.seek(0)
        line = file_handle.readline()
        ttype,wh=line.split()
        self[ttype] = float(wh)

        # Enter Loop
        citer = -1
        while True:
            line = file_handle.readline()
            if line == '':
                break
            ttype,hw = line.split(' ',1)
            if ttype == 'ITER':
                if 'MIN' in hw:
                    break
                citer = citer+1
                self[ttype][citer] = int(hw)
                continue
            else:
                h,w = hw.split()
                h = int(h)
                w = int(w)
                line = file_handle.readline()
            if ttype not in self:
                self[ttype]=np.ndarray((niter,h,w))
            for i in range(h):
                line = file_handle.readline()
                val = np.fromstring(line,sep=' ')
                self[ttype][citer,i,:] = val       
        file_handle.close()
        for item in list(self):
            # print(item)
            if 'VERSION' == item:
                continue
            elif 'ITER' == item:
                continue
            elif item in ['ASPECT','ASPECT_MAX','BETA','CURTOR','PHIEDGE', \
                        'VOLUME','WP','RBTOR','R0','Z0','BETATOR','BETAPOL']:
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
            elif item == 'BALLOON':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_grate'] = np.array(self[item][:,:,3])
                self[item+'_theta'] = np.array(self[item][:,:,4])
                self[item+'_zeta'] = np.array(self[item][:,:,5])
                self[item+'_k'] = np.array(self[item][:,:,6])
            elif item == 'B_PROBES':
                self[item+'_target'] = np.array(self[item][:,:,4])
                self[item+'_sigma'] = np.array(self[item][:,:,5])
                self[item+'_equil'] = np.array(self[item][:,:,6])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_X'] = np.array(self[item][:,:,0])
                self[item+'_Y'] = np.array(self[item][:,:,1])
                self[item+'_Z'] = np.array(self[item][:,:,2])
                self[item+'_MODB'] = np.array(self[item][:,:,3])
            elif item in ['FLUXLOOPS','SEGROG']:
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
            elif item == 'EXTCUR':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_dex'] = np.array(self[item][:,:,3])
            elif item in ['SEPARATRIX','LIMITER']:
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_R'] = np.array(self[item][:,:,4])
                self[item+'_PHI'] = np.array(self[item][:,:,5])
                self[item+'_Z'] = np.array(self[item][:,:,6])
            elif item in ['TI','TE','IOTA','VPHI','PRESS','NE']:
                self[item+'_target'] = np.array(self[item][:,:,4])
                self[item+'_sigma'] = np.array(self[item][:,:,5])
                self[item+'_equil'] = np.array(self[item][:,:,6])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_R'] = np.array(self[item][:,:,0])
                self[item+'_PHI'] = np.array(self[item][:,:,1])
                self[item+'_Z'] = np.array(self[item][:,:,2])
                self[item+'_S'] = np.array(self[item][:,:,3])
            elif item in ['NELINE','FARADAY','SXR']:
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_R0'] = np.array(self[item][:,:,3])
                self[item+'_PHI0'] = np.array(self[item][:,:,4])
                self[item+'_Z0'] = np.array(self[item][:,:,5])
                self[item+'_R1'] = np.array(self[item][:,:,6])
                self[item+'_PHI1'] = np.array(self[item][:,:,7])
                self[item+'_Z1'] = np.array(self[item][:,:,8])
            elif item == 'MSE':
                self[item+'_target'] = np.array(self[item][:,:,4])
                self[item+'_sigma'] = np.array(self[item][:,:,5])
                self[item+'_equil'] = np.array(self[item][:,:,8])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_R'] = np.array(self[item][:,:,0])
                self[item+'_PHI'] = np.array(self[item][:,:,1])
                self[item+'_Z'] = np.array(self[item][:,:,2])
                self[item+'_S'] = np.array(self[item][:,:,3])
                self[item+'_ER'] = np.array(self[item][:,:,6])
                self[item+'_EZ'] = np.array(self[item][:,:,7])
            elif item == 'BOOTSTRAP':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
                self[item+'_avg_jdotb'] = np.array(self[item][:,:,4])
                self[item+'_beam_jdotb'] = np.array(self[item][:,:,5])
                self[item+'_boot_jdotb'] = np.array(self[item][:,:,6])
                self[item+'_jBbs'] = np.array(self[item][:,:,7])
                self[item+'_facnu'] = np.array(self[item][:,:,8])
                self[item+'_bsnorm'] = np.array(self[item][:,:,9])
            elif item == 'HELICITY':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_bnorm'] = np.array(self[item][:,:,3])
            elif item ==  'HELICITY_FULL':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_bnorm'] = np.array(self[item][:,:,3])
                self[item+'_k'] = np.array(self[item][:,:,4])
                self[item+'_m'] = np.array(self[item][:,:,5])
                self[item+'_n'] = np.array(self[item][:,:,6])               
            elif item == 'TXPORT':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
            elif item == 'COIL_BNORM':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_U'] = np.array(self[item][:,:,3])
                self[item+'_V'] = np.array(self[item][:,:,4])
                self[item+'_BNEQ'] = np.array(self[item][:,:,5])
                self[item+'_BNF'] = np.array(self[item][:,:,6])
            elif item == 'ORBIT':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
            elif item == 'J_STAR':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_AVGJSTAR'] = np.array(self[item][:,:,3])
                self[item+'_TRAPSJSTAR'] = np.array(self[item][:,:,4])
                self[item+'_UJSTAR'] = np.array(self[item][:,:,5])
                self[item+'_K'] = np.array(self[item][:,:,6])
                self[item+'_IJSTAR'] = np.array(self[item][:,:,7])
            elif item == 'NEO':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_K'] = np.array(self[item][:,:,3])
            elif item == 'JDOTB':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
            elif item == 'JTOR':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
            elif item == 'DKES':
                self[item+'_target'] = np.array(self[item][:,:,0])
                self[item+'_sigma'] = np.array(self[item][:,:,1])
                self[item+'_equil'] = np.array(self[item][:,:,2])
                self[item+'_chisq'] = ((self[item+'_target'] - self[item+'_equil'])/self[item+'_sigma'])**2
                self[item+'_S'] = np.array(self[item][:,:,3])
                self[item+'_NU'] = np.array(self[item][:,:,4])
                self[item+'_ER'] = np.array(self[item][:,:,5])
                self[item+'_L11P'] = np.array(self[item][:,:,6])
                self[item+'_L11M'] = np.array(self[item][:,:,7])
                self[item+'_L33P'] = np.array(self[item][:,:,8])
                self[item+'_L33M'] = np.array(self[item][:,:,9])
                self[item+'_L31P'] = np.array(self[item][:,:,10])
                self[item+'_L31M'] = np.array(self[item][:,:,11])
                self[item+'_SCAL11'] = np.array(self[item][:,:,12])
                self[item+'_SCAL33'] = np.array(self[item][:,:,13])
                self[item+'_SCAL31'] = np.array(self[item][:,:,14])
        self['chisq'] = ((self['TARGETS'] - self['VALS'])/self['SIGMAS'])**2 
        return

    def plot(self, ax=None, all=True, **kwargs):
        import itertools
        marker = itertools.cycle(('s', '+', '^', 'o', '*')) 
        # plot the overall iterations
        # set default plotting parameters
        if kwargs.get('linewidth') == None:
            kwargs.update({'linewidth': 2.0}) # prefer thicker lines
        if kwargs.get('label') == None:
            kwargs.update({'label': 'Total_Chisq'}) # default label 
        f, ax = get_figure(ax)
        ax.semilogy(self['ITER'], np.sum(self['chisq'], axis=1), **kwargs)
        if all:
            for key in self.keys():
                if '_chisq' in key:
                    label = key.replace('_chisq', '')
                    ax.semilogy(self['ITER'], np.sum(self[key], axis=1), label=label, marker=next(marker), linestyle='')
            plt.legend()
        plt.xlabel('iterations')
        plt.ylabel('chisq')
        return ax

    def plot_helicity(self, it=-1, ordering=0, mn=(None, None), ax=None, **kwargs):     
        """Plot |B| components in Boozer coordinates from BOOZ_XFORM

        Args:
            it (int, optional): Iteration index to be plotted. Defaults to -1.
            ordering (integer, optional): Plot the leading Nordering asymmetric modes. Defaults to 0.
            mn (tuple, optional): Plot the particular (m,n) mode. Defaults to (None, None).
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.
        
        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """ 
        from .booz_xform import BOOZ_XFORM    
        xs = self['HELICITY_FULL_k']
        xs = np.array(np.unique(xs), dtype=int)
        ns = len(xs)
        xx = (xs-1)/np.max(xs) # max(xs) might be different from NS
        vals = np.reshape(self['HELICITY_FULL_equil'][it], (ns, -1))
        xm = np.reshape(np.array(self['HELICITY_FULL_m'][it], dtype=int), (ns, -1))
        xn = np.reshape(np.array(self['HELICITY_FULL_n'][it], dtype=int), (ns, -1))
        return BOOZ_XFORM.plot_helicity(vals, xm[0,:], xn[0,:], xx, ordering, mn, ax, **kwargs)

    def plot_balloon(self, it=-1, ax=None, **kwargs):
        fig, ax = get_figure(ax)
        ax.plot(self['BALLOON_k'][it],self['BALLOON_grate'][it],'o',fillstyle='none')
        ax.set_xlabel('Radial Grid')
        ax.set_ylabel('Growth Rate')
        ax.set_title('COBRA Ballooning Stability (<0 Stable)')
        return ax


import numpy as np 
import matplotlib.pyplot as plt
from .sortedDict import SortedDict
from .misc import trig2real

__all__ = ['BOOZ_XFORM']

class BOOZ_XFORM(SortedDict):
    def __init__(self, filename, **kwargs):
        import xarray
        SortedDict.__init__(self)
        self.output = xarray.open_dataset(filename)
        self.ns = int(self.output['ns_b'].values)
        self.nfp = int(self.output['nfp_b'].values)
        self.xm = self.output['ixm_b'].values
        self.xn = self.output['ixn_b'].values / self.nfp
        self.jlist = self.output['jlist'].values
        self.bmnc = self.output['bmnc_b'].values
        return
    
    @staticmethod
    def write_input(extension, mbooz, nbooz, surfaces):
        """Write a BOO_XFORM input file

        Args:
            extension (str): VMEC output extension.
            mbooz (int): Maximum poloidal mode. 
            nbooz (int): Maximum toroidal mode 
            surfaces (list): List of flux surfaces.
        """        
        template = """{mbooz:d}  {nbooz:d}
'{extension:s}'
{surfaces:s}
        """
        assert mbooz>0, "mbooz should be >0."
        assert nbooz>0, "nbooz should be >0."
        surfaces = ['{:} '.format(i) for i in surfaces]
        surfaces = ''.join(surfaces)
        with open('inbooz.'+extension, 'w') as f:
            f.write(template.format(mbooz=mbooz,
                                    nbooz=nbooz,
                                    extension=extension,
                                    surfaces=surfaces))
        return

    @staticmethod
    def from_vmec(wout, mbooz=0, nbooz=0, surfaces=[]):
        """Prepare BOOZ_XFORM input file from VMEC wout file

        Args:
            wout (str): VMEC wout filename.
            mbooz (int, optional): Maximum poloildal mode number. Defaults to 0 (4*Mpol).
            nbooz (int, optional): Maximum toroidal mode number. Defaults to 0 (4*Ntor).
            surfaces (list, optional): Flux surfaces list. Defaults to [] (1:NS).

        """        
        import xarray
        vmec = xarray.open_dataset(wout, 'r')
        ind = wout.index('wout_') + 5
        end = wout.index('.nc')
        extension = wout[ind:end]
        if mbooz == 0:
            mbooz = 4*int(vmec['mpol'].values) + 1
        if nbooz == 0:
            nbooz = 4*int(vmec['ntor'].values)
        if len(surfaces) == 0:
            ns = int(vmec['ns'].values)
            surfaces = 1 + np.arange(ns)
        vmec.close()
        return BOOZ_XFORM.write_input(extension, mbooz, nbooz, surfaces)
    
    def plot(self, ordering=0, mn=(None, None), ax=None, **kwargs):
        """Plot |B| components 1D profile

        Args:
            ordering (integer, optional): Plot the leading Nordering asymmetric modes. Defaults to 0.
            mn (tuple, optional): Plot the particular (m,n) mode. Defaults to (None, None).
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            kwargs (dict): Keyword arguments for matplotlib.pyplot.plot. Defaults to {}.
        
        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """ 
        xx = self.jlist / self.ns
        return self.plot_helicity(self.bmnc, self.xm, self.xn, xx, ordering, mn, ax, **kwargs)

    @staticmethod
    def plot_helicity(*args, **kwargs):
        vals, xm, xn, xx, ordering, mn, ax = args
        # get figure and ax data
        if ax is None:
            fig, ax = plt.subplots()
        plt.sca(ax)
        # select the top ordering asymmetric terms
        if ordering:
            assert ordering >= 1
            data = np.linalg.norm(vals, axis=0)
            ind_arg = np.argsort(data)
            for i in range(ordering):
                ind = ind_arg[-1-i] # index of the i-th largest term
                m = xm[ind]
                n = xn[ind]
                kwargs['label'] = 'm={:}, n={:}'.format(m,n)
                ax.semilogy(xx, vals[:, ind], **kwargs)
            ylabel = r'$\frac{B_{m,n}}{ \Vert B_{n=0} \Vert }$'
        else:
            # determine filter condition
            if mn[0] is not None:
                mfilter = (xm == mn[0])
                m = 'm={:}'.format(mn[0])
            else:
                mfilter = np.full(np.shape(xm), True)
                m = 'm'
            if mn[1] is not None:
                nfilter = (xn == mn[1])
                n = 'n={:}'.format(mn[1])
            else:
                nfilter = (xn != 0)
                n = r'n \neq 0'
            cond = np.logical_and(mfilter, nfilter)
            #data = np.reshape(vals[:, cond], (ns, -1))
            data = vals[:, cond]
            line = ax.semilogy(xx, np.linalg.norm(data, axis=1), **kwargs)
            ylabel = r'$ \frac{{ \Vert B_{{ {:},{:} }} \Vert }}{{ \Vert B_{{n=0}} \Vert }} $'.format(m, n)
        plt.xlabel('normalized flux (s)', fontsize=16)
        plt.ylabel(ylabel, fontsize=16)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        return ax

    def plot2d(self, ns=-1, npol=128, ntor=128, ax=None, contour=True, **kwargs):
        """Plot |B| contour on the flux surface

        Args:
            ns (int, optional): Flux surface index in self.jlist. Defaults to -1.
            npol (int, optional): Poloidal resolution. Defaults to 128.
            ntor (int, optional): Toroidal resolution. Defaults to 128.
            ax (Matplotlib axis, optional): Matplotlib axis to be plotted on. Defaults to None.
            contour (bool, optional): Whether plotting the contour lines. Defaults to True.

        Returns:
            ax (Matplotlib axis): Matplotlib axis plotted on.
        """        
        _theta = np.linspace(0, np.pi*2, npol, endpoint=True)
        _zeta  = np.linspace(0, np.pi*2, ntor, endpoint=True)
        modB = trig2real(_theta, _zeta, self.xm, self.xn, self.bmnc[ns,:])
        # get figure and ax data
        if ax is None:
            fig, ax = plt.subplots()
        plt.sca(ax)
        bl = ax.imshow(modB, origin='lower', extent=[0, 2*np.pi, 0, 2*np.pi])
        plt.xticks(np.linspace(0, 2*np.pi, 3), 
                   ['0', r'$\pi/{:d}$'.format(self.nfp), r'$2\pi/{:d}$'.format(self.nfp)], fontsize=15)
        plt.yticks(np.linspace(0, 2*np.pi, 3), ['0', r'$\pi$', r'$2\pi$'], fontsize=15)
        if contour:
            ax.contour(_zeta, _theta, modB, colors='grey')
        clb = plt.colorbar(bl)
        clb.ax.tick_params(labelsize=15)
        clb.ax.set_title('|B|(T)', fontsize=15)
        plt.xlabel(r'$\zeta_{Boozer}$', fontsize=16)
        plt.ylabel(r'$\theta_{Boozer}$', fontsize=16)
        import matplotlib.ticker as tck
        ax.xaxis.set_minor_locator(tck.AutoMinorLocator())
        ax.yaxis.set_minor_locator(tck.AutoMinorLocator())
        return ax

    def to_FOCUS(self, ns=-1, focus_file='plasma.boundary', tol=1.0E-8):
        """Write a FOCUS plasma boundary file in Boozer coordinates

        Args:
            ns (int, optional): Flux surface label to be exported. Defaults to -1.
            focus_file (str, optional): FOCUS plasma boundary file name. Defaults to 'plasma.boundary'.
            tol (float, optional): Truncated tolerance. Defaults to 1.0E-8.
        """        
        mn = int(self.output['mnboz_b'].values)
        rbc = np.array(self.output['rmnc_b'][ns,:])
        #rbs = np.zeros(mn)
        zbs = np.array(self.output['zmns_b'][ns,:])
        #zbc = np.zeros(mn)
        pmns = np.array(self.output['pmns_b'][ns,:])
        #pmnc = np.zeros(mn)
        Nbnf = 0
        # count non-zero terms
        amn = 0
        for imn in range(mn):
            if (abs(rbc[imn])+abs(zbs[imn]+abs(pmns[imn]))) > tol :
                amn += 1 # number of nonzero coef.
        # write FOCUS plasma boundary file
        with open(focus_file, 'w') as fofile:
            fofile.write('# bmn   bNfp   nbf '+'\n')
            fofile.write("{:d} \t {:d} \t {:d} \n".format(amn, self.nfp, Nbnf))
            fofile.write('#plasma boundary'+'\n')
            fofile.write('# n m Rbc Rbs Zbc Zbs Pmnc Pmns'+'\n')
            for imn in range(mn):
                if (abs(rbc[imn])+abs(zbs[imn]+abs(pmns[imn]))) > tol :
                    fofile.write("{:4d}  {:4d} \t {:23.15E}  {:12.5E}  {:12.5E} {:23.15E} {:12.5E}  {:23.15E} \n"
                                .format(self.xn[imn], self.xm[imn], rbc[imn], 0.0, 0.0, zbs[imn], 0.0, pmns[imn]))
            fofile.write("#Bn harmonics \n")
            fofile.write('# n m bnc bns'+'\n')          
        return



import numpy as np
u0_d_4pi = 1.0E-7
class SingleCoil(object):
    '''
    Single coil class represented as discrete points in Cartesian coordinates
    '''
    def __init__(self, x=[], y=[], z=[], I=0.0, name='coil1', group=1):
        assert len(x) == len(y) == len(z), "dimension not consistent"
        self.x = np.asarray(x)
        self.y = np.asarray(y)
        self.z = np.asarray(z)
        self.I = I
        self.name = name
        self.group = group
        self.xt = None
        self.yt = None
        self.zt = None
        return
    
    def __del__(self):
        class_name = self.__class__.__name__
        #print(class_name, "destroyed")

    def plot(self, engine='mayavi', **kwargs):
        '''
        plot coil as a line
        '''
        if engine == 'pyplot':
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            # plot in matplotlib.pyplot
            if plt.get_fignums():
                fig = plt.gcf()
                ax = plt.gca()
            else :
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
            ax.plot(self.x, self.y, self.z, **kwargs)
        elif engine == 'mayavi':
            # plot 3D line in mayavi.mlab
            from mayavi import mlab # to overrid plt.mlab
            mlab.plot3d(self.x, self.y, self.z, **kwargs)
        else:
            raise ValueError('Invalid engine option {pyplot, mayavi, noplot}')
        return        

    def rectangle(self, width=0.1, height=0.1, frame='centroid', **kwargs):
        '''
        This function expand single coil filament to a rectangle coil;
        
        width
        height
        winding : winding surface data;
        tol: root find tolarence
        '''
        n = np.size(self.x)
        dt = 2*np.pi/(n-1)
        # calculate the tangent 
        if self.xt is None:
            self.spline_tangent()
        xt = self.xt
        yt = self.yt
        zt = self.zt
        # xt = np.gradient(self.x)/dt
        # yt = np.gradient(self.y)/dt
        # zt = np.gradient(self.z)/dt
        tt = np.sqrt(xt*xt + yt*yt + zt*zt)
        xt = xt/tt
        yt = yt/tt
        zt = zt/tt

        # use surface normal if needed
        if frame == 'centroid':
            # use the geometry center is a good idea
            center_x = np.average(self.x[0:n-1])
            center_y = np.average(self.y[0:n-1])
            center_z = np.average(self.z[0:n-1])
            xn = self.x - center_x
            yn = self.y - center_y
            zn = self.z - center_z
            nt = xn*xt + yn*yt + zn*zt
            xn = xn - nt*xt
            yn = yn - nt*yt
            zn = zn - nt*zt
        elif frame == 'frenet':
            self.spline_tangent(der=2)
            xn = self.xa
            yn = self.ya
            zn = self.za
        elif frame == 'parallel':
            # parallel transport frame 
            # Hanson & Ma, Parallel Transp ort Approach to Curve Framing, 1995
            def rotate(x, ang):
                c = np.cos(ang)
                s = np.sin(ang)
                return [[c + x[0]**2*(1-c), x[0]*x[1]*(1-c)-s*x[2], x[2]*x[0]*(1-c)+s*x[1]],
                        [x[0]*x[1]*(1-c)+s*x[2], c+x[1]**2*(1-c), x[2]*x[1]*(1-c)-s*x[0]],
                        [x[0]*x[2]*(1-c)-s*x[1], x[1]*x[2]*(1-c)+s*x[0], c+x[2]**2*(1-c)]]
            T = np.transpose([self.xt, self.yt, self.zt])
            T = T/np.linalg.norm(T, axis=1)[:, np.newaxis]
            B = np.cross(T[:-1], T[1:], axis=1)
            B = B/np.linalg.norm(B, axis=1)[:, np.newaxis]
            theta = np.arccos(np.sum(T[:-1]*T[1:], axis=1))
            V = np.zeros_like(T)
            vx = kwargs['vx']
            vy = kwargs['vy']
            vz = -(vx*T[0,0] + vy*T[0,1])/T[0,2]
            vv = np.linalg.norm([vx, vy, vz])
            V[0,:] = [vx/vv, vy/vv, vz/vv]
            print(np.dot(V[0,:], T[0,:]))
            for i in range(len(theta)):
                V[i+1,:] = rotate(B[i,:], theta[i]) @ V[i,:]
            xn = V[:,0]
            yn = V[:,1]
            zn = V[:,2]
        else:
            assert True, "not finished"
        
        nn = np.sqrt(xn*xn + yn*yn + zn*zn)
        xn = xn/nn
        yn = yn/nn
        zn = zn/nn
        # calculate the bi-normal
        xb = yt*zn - yn*zt
        yb = zt*xn - zn*xt
        zb = xt*yn - xn*yt
        bb = np.sqrt(xb*xb + yb*yb + zb*zb)
        xb = xb/bb
        yb = yb/bb
        zb = zb/bb
        # get the boundary lines
        z1 = self.z - width/2*zb + height/2*zn
        x1 = self.x - width/2*xb + height/2*xn
        x2 = self.x + width/2*xb + height/2*xn
        y2 = self.y + width/2*yb + height/2*yn
        z2 = self.z + width/2*zb + height/2*zn
        x3 = self.x + width/2*xb - height/2*xn
        y3 = self.y + width/2*yb - height/2*yn
        z3 = self.z + width/2*zb - height/2*zn
        x4 = self.x - width/2*xb - height/2*xn
        y4 = self.y - width/2*yb - height/2*yn
        z4 = self.z - width/2*zb - height/2*zn
        y1 = self.y - width/2*yb + height/2*yn
        # assemble 
        xx = np.array([x1, x2, x3, x4, x1])
        yy = np.array([y1, y2, y3, y4, y1])
        zz = np.array([z1, z2, z3, z4, z1])

        return xx, yy, zz
        
    def interpolate(self, num=256):
        '''
        Interpolate to increase more data points
        '''
        from scipy.interpolate import interp1d
        cur_len = len(self.x)
        assert cur_len > 0
        theta = np.linspace(0, 1, num=cur_len, endpoint=True)
        theta_new = np.linspace(0, 1, num=num, endpoint=True)
        for xyz in [self.x, self.y, self.z]:
            f = interp1d(theta, xyz, kind='cubic')
            xyz[:] = f(theta_new)
        return
    
    def magnify(self, ratio):
        """
        magnify coil with a ratio
        """
        # number of points
        nseg = len(self.x)
        # assuming closed curve; should be revised
        if True: #abs(self.x[0] - self.x[-1]) < 1.0E-8:
            nseg -= 1
        assert nseg>1
        # get centroid
        centroid = np.array([np.sum(self.x[0:nseg])/nseg, 
                             np.sum(self.y[0:nseg])/nseg,
                             np.sum(self.z[0:nseg])/nseg])
        # magnify
        for i in range(nseg):
            xyz = np.array([self.x[i], self.y[i], self.z[i]])
            dr = xyz-centroid
            [self.x[i], self.y[i], self.z[i]] = centroid + ratio*dr
        try:
            self.x[nseg] = self.x[0]
            self.y[nseg] = self.y[0]
            self.z[nseg] = self.z[0]
            return
        except:
            return

    def bfield_HH(self, pos, **kwargs):
        """Calculate B field at an arbitrary point using the Hanson-Hirshman expression

        Arguments:
            pos {array-like} -- Cartesian coordinates for the evaluation point

        Returns:
            ndarray -- the calculated magnetic field vector
        """        
        xyz = np.array([self.x, self.y, self.z]).T
        pos = np.atleast_2d(pos)
        assert (pos.shape)[1]  == 3
        Rvec = pos[:,np.newaxis,:] - xyz[np.newaxis,:,:]
        assert (Rvec.shape)[-1] == 3
        RR = np.linalg.norm(Rvec, axis=2)
        Riv = Rvec[:, :-1, :]
        Rfv = Rvec[:, 1:, :]
        Ri = RR[:,:-1]
        Rf = RR[:,1:]
        B = np.sum(np.cross(Riv, Rfv)*((Ri+Rf)/((Ri*Rf)*(Ri*Rf+np.sum(Riv*Rfv, axis=2))))[:, :, np.newaxis], axis=1)\
            *u0_d_4pi*self.I
        return B

    def bfield(self, pos):        
        ob_pos = np.atleast_1d(pos)
        dx = ob_pos[0] - self.x[:-1]
        dy = ob_pos[1] - self.y[:-1]
        dz = ob_pos[2] - self.z[:-1]
        dr = dx*dx + dy*dy +dz*dz
        Bx = (dz*self.yt[:-1] - dy*self.zt[:-1])*np.power(dr, -1.5)
        By = (dx*self.zt[:-1] - dz*self.xt[:-1])*np.power(dr, -1.5)
        Bz = (dy*self.xt[:-1] - dx*self.yt[:-1])*np.power(dr, -1.5)
        B = np.array([np.sum(Bx), np.sum(By), np.sum(Bz)])*self.dt*u0_d_4pi*self.I
        return B        
    
    def spline_tangent(self, order=3, der=1):
        """Calculate the tangent of coil using spline interpolation

        Args:
            order (int, optional): Order of spline interpolation used. Defaults to 3.
        """        
        from scipy import interpolate
        t = np.linspace(0, 2*np.pi, len(self.x), endpoint=True)
        self.dt = 2*np.pi/(len(self.x)-1)
        fx = interpolate.splrep(t, self.x, s=0, k=order)
        fy = interpolate.splrep(t, self.y, s=0, k=order)
        fz = interpolate.splrep(t, self.z, s=0, k=order)
        self.xt = interpolate.splev(t, fx, der=1)
        self.yt = interpolate.splev(t, fy, der=1)
        self.zt = interpolate.splev(t, fz, der=1)
        if der == 2:
            self.xa = interpolate.splev(t, fx, der=2)
            self.ya = interpolate.splev(t, fy, der=2)
            self.za = interpolate.splev(t, fz, der=2)
        return 

    def fourier_tangent(self):
        from .misc import fft_deriv
        self.dt = 2*np.pi/(len(self.x)-1)
        fftxy = fft_deriv(self.x[:-1]+1j*self.y[:-1])
        fftz = fft_deriv(self.z[:-1])
        self.xt = np.real(fftxy)
        self.yt = np.imag(fftxy)
        self.zt = np.real(fftz)
        self.xt = np.concatenate((self.xt, self.xt[0:1]))
        self.yt = np.concatenate((self.yt, self.yt[0:1]))
        self.zt = np.concatenate((self.zt, self.zt[0:1]))
        return         

    def toVTK(self, vtkname, **kwargs):
        """Write a VTK file

        Args:
            vtkname (string): VTK filename
        """        
        from pyevtk.hl import polyLinesToVTK
        kwargs.setdefault('cellData', {})
        kwargs['cellData'].setdefault('I', np.array(elf.I))
        polyLinesToVTK(vtkname, np.array(self.x), np.array(self.y), np.array(self.z),
                       np.array([len(self.x)]), **kwargs)
        return
        
class Coil(object):
    def __init__(self, xx=[], yy=[], zz=[], II=[], names=[], groups=[]):
        assert len(xx) == len(yy) == len(zz) == len(II) == len(names) == len(groups), "dimension not consistent"
        self.num = len(xx)
        self.data = []
        for i in range(self.num):
            self.data.append(SingleCoil(x=xx[i], y=yy[i], z=zz[i], I=II[i], name=names[i], group=groups[i]))
        self.index = 0
        return 

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < self.num:
            self.index += 1
            return self.data[self.index-1]
        else:
            self.index = 0 
            raise StopIteration()

    def __len__(self):
        return self.num

    @classmethod
    def read_makegrid(cls, filename):
        """read MAKEGRID format 
        Args:
            filename (str) : file path and name
        Returns:
            Coil class
        """
        import os
        # check existence
        if not os.path.exists(filename) :
            raise IOError ("File not existed. Please check again!")
        # read and parse data
        cls.header = ''
        with open(filename,'r') as coilfile: #read coil xyz and I
            cls.header = ''.join((coilfile.readline(), coilfile.readline(), coilfile.readline()))
            icoil = 0
            xx = [[]]; yy = [[]]; zz = [[]]
            II = []; names = []; groups = []
            tmpI = 0.0
            for line in coilfile:
                linelist = line.split()
                if len(linelist) < 4 :
                    #print("End of file or invalid format!")                    
                    break
                xx[icoil].append(float(linelist[0]))
                yy[icoil].append(float(linelist[1]))
                zz[icoil].append(float(linelist[2]))
                if len(linelist) == 4 :
                    tmpI = float(linelist[-1])
                if len(linelist) > 4 :
                    II.append(tmpI)
                    try:
                        group = int(linelist[4])
                    except:
                        group = len(groups) + 1
                    groups.append(group)
                    try:
                        name = linelist[5]
                    except IndexError:
                        name = 'coil'
                    names.append(name)
                    icoil = icoil + 1
                    xx.append([]); yy.append([]); zz.append([])
        xx.pop(); yy.pop(); zz.pop()
        # print(len(xx) , len(yy) , len(zz) , len(II) , len(names) , len(groups))
        return cls(xx=xx, yy=yy, zz=zz, II=II, names=names, groups=groups)

    def plot(self, irange=[], **kwargs):
        """Plot coils in mayavi or matplotlib  

        Args:
            irange (list, optional): Coil list to be plotted. Defaults to [].
        """        
        if len(irange)==0:
            irange = range(self.num)    
        for i in irange:
            self.data[i].plot(**kwargs)
        return

    def save_makegrid(self, filename, nfp=1, **kwargs):
        """write in MAKEGRID format

        Arguments:
            filename {str} -- file name and path
            nfp {int} -- number of toroidal periodicity (default: 1)
        """
        assert len(self) > 0
        with open(filename, 'w') as wfile :
            wfile.write("periods {:3d} \n".format(nfp))
            wfile.write("begin filament \n")
            wfile.write("mirror NIL \n")
            for icoil in list(self):
                Nseg = len(icoil.x) # number of segments;
                assert Nseg > 1
                for iseg in range(Nseg-1): # the last point match the first one;
                    wfile.write("{:15.7E} {:15.7E} {:15.7E} {:15.7E}\n".format(
                        icoil.x[iseg], icoil.y[iseg], icoil.z[iseg], icoil.I))
                wfile.write("{:15.7E} {:15.7E} {:15.7E} {:15.7E} {:} {:10} \n".format(
                    icoil.x[0], icoil.y[0], icoil.z[0], 0.0, icoil.group, icoil.name))
            wfile.write("end \n")
        return

    def toVTK(self, vtkname, **kwargs):
        """Write entire coil set into a VTK file

        Args:
            vtkname (str): VTK filename
            kwargs (dict): Optional kwargs passed to "polyLinesToVTK"
        """        
        from pyevtk.hl import polyLinesToVTK
        currents = []
        groups = []
        x = []
        y = []
        z = []
        lx = []
        for icoil in list(self):
            currents.append(icoil.I)
            groups.append(icoil.group)
            x.append(icoil.x)
            y.append(icoil.y)
            z.append(icoil.z)
            lx.append(len(icoil.x))
        kwargs.setdefault('cellData', {})
        kwargs['cellData'].setdefault('I', np.array(currents))
        kwargs['cellData'].setdefault('Igroup', np.array(groups))
        polyLinesToVTK(vtkname, np.concatenate(x), np.concatenate(y), np.concatenate(z),
                        np.array(lx), **kwargs)
        return

    def __del__(self):
        class_name = self.__class__.__name__



import numpy as np
from .misc import xy2rp, map_matrix, toroidal_period

class Dipole(object):
    '''
    magnetic dipole class
    '''
    def __init__(self, **kwargs):
        '''
        Initialize empty class
        '''
        self.num = 0 # total number of dipoles
        self.nfp = 1 # toroidal_periodicity
        self.momentq = 1 # q expotent
        self.sp_switch = False # switch to indicate if using spherical coordinates
        self.xyz_switch = False # switch to indicate if using spherical coordinates
        self.old = False # old format or new
        #self.symmetry = 2 # 0: no symmetry; 1: periodicity; 2: stellarator symmetry (+periodicity)
        if 'mm' in kwargs: # spherical coord
            self.ox = kwargs['ox']
            self.oy = kwargs['oy']
            self.oz = kwargs['oz']
            self.mm = kwargs['mm']
            self.mt = kwargs['mt']
            self.mp = kwargs['mp']
            self.pho = kwargs['pho']
            self.momentq = kwargs['momentq']
            self.Ic = kwargs['Ic']
            self.Lc = kwargs['Lc']
            self.num = len(self.ox)
            self.sp_switch = True
            self.filename = kwargs.get('filename', 'dipole')
            self.name = kwargs.get('name', ['pm_{:010d}'.format(i) for i in range(1, self.num+1)])
            self.symm = kwargs.get('symm', 2*np.ones(self.num, dtype=int))
            self.rho = self.pho**self.momentq
        elif 'mx' in kwargs: # cartesian coord
            self.ox = kwargs['ox']
            self.oy = kwargs['oy']
            self.oz = kwargs['oz']
            self.mx = kwargs['mx']
            self.my = kwargs['my']
            self.mz = kwargs['mz']
            self.mm = np.sqrt(self.mx**2 + self.my**2 + self.mz**2)
            self.num = len(self.ox)
            self.pho = np.ones(self.num)
            self.Ic = np.zeros(self.num, dtype=int)
            self.Lc = np.zeros(self.num, dtype=int)        
            self.xyz_switch = True
            self.filename = kwargs.get('filename', 'dipole')
            self.name = kwargs.get('name', ['pm_{:010d}'.format(i) for i in range(1, self.num+1)])
            self.symm = kwargs.get('symm', 2*np.ones(self.num, dtype=int))
            self.rho = np.ones(self.num)
        return

    @classmethod
    def open(cls, filename, verbose=False, **kwargs):
        """Read FAMUS dipoles

        Args:
            filename (str): path to open the file.
            verbose (bool, optional): Whether to print additional info. Defaults to False.

        Returns:
            Dipole: Dipole class
        """        
        import pandas as pd
        with open(filename, 'r') as coilfile:
            coilfile.readline()
            line = coilfile.readline().split()
            try:
                momentq = int(line[1])
            except:
                print("Moment Q factor was not read. Default=1.")
                momentq = 1
        data = pd.read_csv(filename, skiprows=3, header=None)
        symm = np.array(data[1], dtype=int)
        name = np.array(data[2]) #type string       
        ox = np.array(data[3], dtype=float)
        oy = np.array(data[4], dtype=float)
        oz = np.array(data[5], dtype=float)
        Ic = np.array(data[6], dtype=int)
        mm = np.array(data[7], dtype=float)
        pho= np.array(data[8], dtype=float)
        Lc = np.array(data[9], dtype=int)
        mp = np.array(data[10], dtype=float)
        mt = np.array(data[11], dtype=float)
        if verbose:
            print('Read {:d} dipoles from {:}'.format(len(ox), filename))
        return cls(symm=symm, name=name, ox=ox, oy=oy, oz=oz, 
                   Ic=Ic, mm=mm, Lc=Lc, mp=mp, mt=mt, pho=pho, 
                   momentq=momentq, filename=filename)

    @classmethod
    def read_dipole_old(cls, filename, zeta=0.0, zeta1=np.pi*2, **kwargs):
        """read diploes from FOCUS format (old)

        Args:
            filename (str): path to open the file.
            zeta (float, optional): Starting toroidal angle for clipping. Defaults to 0.0.
            zeta1 (float, optional): [Ending toroidal angle for clipping]. Defaults to np.pi*2.

        Returns:
            Dipole: Dipole class
        """
        symm = []; name = []
        ox = []; oy = []; oz = []
        mm = []; mp = []; mt = []
        Ic = []; Lc = []
        with open(filename, 'r') as coilfile:
            coilfile.readline()
            Ncoils = int(coilfile.readline()[0])
            for icoil in range(Ncoils):
                coilfile.readline()
                coilfile.readline()
                linelist = coilfile.readline().split()
                if int(linelist[0]) == 3 : # central current and background Bz
                    coilfile.readline()
                    linelist = coilfile.readline().split() 
                elif int(linelist[0]) == 2 : # dipoles
                    symm.append(int(linelist[1]))
                    name.append(linelist[2])
                    coilfile.readline()
                    linelist = coilfile.readline().split()
                    r, phi = xy2rp(float(linelist[1]), float(linelist[2]))
                    if phi>=zeta and phi<=zeta1:
                        Lc.append(  int(linelist[0]))
                        ox.append(float(linelist[1]))
                        oy.append(float(linelist[2]))
                        oz.append(float(linelist[3]))
                        Ic.append(  int(linelist[4]))
                        mm.append(float(linelist[5]))
                        mt.append(float(linelist[6]))
                        mp.append(float(linelist[7]))
                elif int(linelist[0]) == 1 : # Fourier coils
                    for i in range(11):
                        coilfile.readline()
                else :
                    raise ValueError('Invalid coiltype = {:d}.'.format(int(linelist[0])))
        nc = len(ox)
        if nc == 0 :
            print('Warning: no dipoles was read from '+filename)
            return
        print('Read {:d} dipoles from {:}. Please manually set self.old=True.'.format(nc, filename))
        symm = np.array(symm)
        ox = np.array(ox)
        oy = np.array(oy)
        oz = np.array(oz)
        mm = np.array(mm)
        mt = np.array(mt)
        mp = np.array(mp)
        Lc = np.array(Lc)
        Ic = np.array(Ic)
        pho = np.ones_like(ox)
        momentq = 1
        return cls( symm=symm, name=name, filename=filename,
                    ox=ox, oy=oy, oz=oz, Ic=Ic, mm=mm, Lc=Lc, 
                    mp=mp, mt=mt, pho=pho, momentq=momentq )

    @classmethod
    def from_regcoil(cls, regcoilname, winding, ilambda=-1, 
                     symmetry='full', num_pol=128, num_tor=128, m0=None, half_shift=True):
        """Initialize from REGCOIL current potential

        Args:
            regcoilname (str): REGCOIL netcdf output file.
            winding (str, optional): NESCOIL input format winding surface. 
            ilambda (int, optional): Lambda index in REGCOIL output. Defaults to -1.
            symmetry (str, optional): Stellarator symmetry option. Defaults to 'full'.
            num_pol (int, optional): Number of poloidal dipoles. Defaults to 128.
            num_tor (int, optional): Number of toroidal dipoles. Defaults to 128.
            m0 ([type], optional): Magnetization limit per thickness. Defaults to None.
            half_shift (bool, optional): Logical flag to determine if half-grid shifted. Defaults to True.

        Returns:
            None
        """                     
        # read regcoil output
        from scipy.io import netcdf
        f = netcdf.netcdf_file(regcoilname,'r',mmap=False)
        nfp = f.variables['nfp'][()]
        ntheta_plasma = f.variables['ntheta_plasma'][()]
        ntheta_coil = f.variables['ntheta_coil'][()]
        nzeta_plasma = f.variables['nzeta_plasma'][()]
        nzeta_coil = f.variables['nzeta_coil'][()]
        nzetal_plasma = f.variables['nzetal_plasma'][()]
        nzetal_coil = f.variables['nzetal_coil'][()]
        theta_plasma = f.variables['theta_plasma'][()]
        theta_coil = f.variables['theta_coil'][()]
        zeta_plasma = f.variables['zeta_plasma'][()]
        zeta_coil = f.variables['zeta_coil'][()]
        zetal_plasma = f.variables['zetal_plasma'][()]
        zetal_coil = f.variables['zetal_coil'][()]
        r_plasma  = f.variables['r_plasma'][()]
        #normal_coil = f.variables['normal_coil'][()]
        r_coil  = f.variables['r_coil'][()]
        xm_coil = f.variables['xm_coil'][()]
        xn_coil = f.variables['xn_coil'][()]
        xm_potential = f.variables['xm_potential'][()]
        xn_potential = f.variables['xn_potential'][()]
        mnmax_coil = f.variables['mnmax_coil'][()]
        chi2_B = f.variables['chi2_B'][()]
        single_valued_current_potential_thetazeta = f.variables['single_valued_current_potential_thetazeta'][()]
        current_potential = f.variables['current_potential'][()]
        Bnormal_from_plasma_current = f.variables['Bnormal_from_plasma_current'][()]
        Bnormal_from_net_coil_currents = f.variables['Bnormal_from_net_coil_currents'][()]
        Bnormal_total = f.variables['Bnormal_total'][()]
        net_poloidal_current_Amperes = f.variables['net_poloidal_current_Amperes'][()]
        phi_mn = f.variables['single_valued_current_potential_mn'][()][ilambda,:]
        symmetry_option = f.variables['symmetry_option'][()]
        f.close()

        mn_max = len(xm_potential)
        phi_sin = np.zeros(mn_max)
        phi_cos = np.zeros(mn_max)
        if symmetry_option == 1:
            phi_sin = phi_mn[0:mn_max]
        elif symmetry_option == 2:
            phi_cos = phi_mn[0:mn_max]
        elif symmetry_option == 3:
            phi_sin = phi_mn[0:mn_max]
            phi_cos = phi_mn[mn_max:]
        else:
            raise ValueError('Something wrong the symmetry_option: {:}'.format(symmetry_option))

        def func_new(theta, zeta):
            """ get current potential of (theta, zeta)

            Parameters:
            theta -- float array_like, poloidal angle
            zeta -- float array_like, toroidal angle value

            Returns:
            phi -- float array_like
            """
            assert len(np.atleast_1d(theta)) == len(np.atleast_1d(zeta)), "theta, zeta should be equal size"
            # mt - nz (in matrix)
            _mtnz = np.matmul( np.reshape(xm_potential, (-1,1)), np.reshape(theta, (1,-1)) ) \
                  - np.matmul( np.reshape(xn_potential, (-1,1)), np.reshape( zeta, (1,-1)) ) 
            _sin = np.sin(_mtnz)
            _cos = np.cos(_mtnz)

            phi = np.matmul( np.reshape(phi_sin, (1,-1)), _sin ) + np.matmul( np.reshape(phi_cos, (1,-1)), _cos ) 
            return phi.ravel()
        
        # initialize dipoles
        num = num_pol * num_tor
        if symmetry.lower() == 'full':
            zeta_end = 2*np.pi
            symm = 0 + np.zeros(num, dtype=np.int)      
        elif symmetry.lower()  == 'one':
            zeta_end = 2*np.pi/nfp
            symm = 1 + np.zeros(num, dtype=np.int)  
        elif symmetry.lower()  == 'half':
            zeta_end = 2*np.pi/nfp/2
            symm = 2 + np.zeros(num, dtype=np.int)            
        else:
            raise ValueError('No such symmetry option!')

        ox = []
        oy = []
        oz = []
        mt = []
        mp = []
        mm = []

        # discretization
        if half_shift: # toroidally shift half grid
            half = zeta_end/(num_tor*2)
        else:
            half = 0
        theta_array = np.linspace(0, 2*np.pi, num_pol, endpoint=False)
        zeta_array = half + np.linspace(0, zeta_end, num_tor, endpoint=False)
        dtheta = 2*np.pi/num_pol
        dzeta = zeta_end/num_tor
        tv, zv = np.meshgrid(theta_array, zeta_array, indexing='ij')

        from .surface import FourSurf
        wind = FourSurf.read_winding_surfce(winding)    
        data = wind.xyz(tv, zv, normal=True)
        phi = func_new(tv, zv)
        ox = data[0]
        oy = data[1]
        oz = data[2]
        n = data[3]
        nn = np.linalg.norm(n, axis=1)
        mt = np.arccos(n[:,2]/nn)
        mp = np.arctan2(n[:,1], n[:,0])
        if m0 is None: # not specify the allowable magnetization
            mm = -phi*nn*dtheta*dzeta
            pho = np.ones_like(mm)
        else:
            mm = m0*nn*dtheta*dzeta
            pho = -phi/m0

        return cls( symm=symm, ox=ox, oy=oy, oz=oz, mp=mp, mt=mt, mm=mm, pho=pho,
                    Ic=np.ones_like(ox, dtype=int), Lc=np.ones_like(ox, dtype=int), momentq=1)

    def sp2xyz(self):
        '''
        spherical coordinates to cartesian coordinates
        '''
        assert self.sp_switch == True, "You are not using spherical coordinates"
        if self.old:
            self.mx = self.mm * np.sin(self.mt) * np.cos(self.mp) 
            self.my = self.mm * np.sin(self.mt) * np.sin(self.mp) 
            self.mz = self.mm * np.cos(self.mt)
        else :
            self.mx = self.mm * self.pho**self.momentq * np.sin(self.mt) * np.cos(self.mp) 
            self.my = self.mm * self.pho**self.momentq * np.sin(self.mt) * np.sin(self.mp) 
            self.mz = self.mm * self.pho**self.momentq * np.cos(self.mt)
        self.xyz_switch = True
        return

    def xyz2sp(self):
        '''
        cartesian coordinates to spherical coordinates
        '''
        assert self.xyz_switch == True, "You are not using cartesian coordinates"
        if self.old :
            self.mm = np.sqrt(self.mx*self.mx + self.my*self.my + self.mz*self.mz)
            #self.pho = np.ones_like(self.mm)
            #self.momentq = 1
        else :
            self.pho = np.power(np.sqrt(self.mx*self.mx + self.my*self.my + self.mz*self.mz)/self.mm, 1.0/self.momentq)
        self.mp = np.arctan2(self.my, self.mx)
        self.mt = np.arccos(self.mz/(self.mm*self.pho**self.momentq))
        self.sp_switch = True
        return   

    def save(self, filename, unique=False):
        '''
        write diploes from FOCUS format
        '''
        if not self.sp_switch:
            self.xyz2sp()
        with open(filename, 'w') as wfile :
            wfile.write(" # Total number of dipoles,  momentq \n")
            if unique:
                wfile.write("{:6d},  {:4d}\n".format(self.num/self.nfp, self.momentq))
            else :
                wfile.write("{:6d},  {:4d}\n".format(self.num, self.momentq))
            if self.old:
                for icoil in range(self.num):
                    if unique:
                        if np.mod(icoil, self.nfp)==0:
                            continue
                    wfile.write("#-----------------{}---------------------------\n".format(icoil+1))
                    wfile.write("#coil_type   symm  coil_name \n")
                    wfile.write("   {:1d}  {:1d}  {:} \n".format(2, self.symm[icoil], self.name[icoil]))
                    wfile.write("#  Lc  ox   oy   oz  Ic  I  mp  mt \n")
                    wfile.write("{:6d} {:23.15E} {:23.15E} {:23.15E} {:6d} {:23.15E} {:23.15E} {:23.15E}\n"\
                       .format(self.Lc[icoil], self.ox[icoil], self.oy[icoil], self.oz[icoil], \
                                   self.Ic[icoil], self.mm[icoil], self.mt[icoil], self.mp[icoil] ))
            else:
                wfile.write('#coiltype, symmetry,  coilname,  ox,  oy,  oz,  Ic,  M_0,  pho,  Lc,  mp,  mt \n')
                for i in range(self.num):
                    if unique:
                        if np.mod(i, self.nfp)==0:
                            continue
                    wfile.write(" 2, {:1d}, {:}, {:15.8E}, {:15.8E}, {:15.8E}, {:2d}, {:15.8E}," \
                   "{:15.8E}, {:2d}, {:15.8E}, {:15.8E} \n".format(self.symm[i], self.name[i], self.ox[i], self.oy[i], self.oz[i], 
                   self.Ic[i], self.mm[i], self.pho[i], self.Lc[i], self.mp[i], self.mt[i]))
        return
    
    def toVTK(self, vtkname, dim=(1), close=False, ntnz=False, toroidal=False, **kwargs):
        """write dipole data into a VTK file

        Args:
            vtkname (str): VTK filename, will be appended with .vts or .vtu. 
            dim (tuple, optional): Dimension information if saved as structured grids. Defaults to (1).
            close (bool, optional): Logical flag to manually close the gaps. Defaults to False.
            ntnz (bool, optional): Logical flag of theta-zeta order. Defaults to False.
            toroidal (bool, optional): Logical flag of filling in the toroidal gap. Defaults to False.

        Returns:
            None
        """        
        from pyevtk.hl import gridToVTK, pointsToVTK
        if not self.xyz_switch:
            self.sp2xyz()
        if vtkname is None:
            vtkname = self.filename
        dim = np.atleast_1d(dim)
        if len(dim) == 1: # save as points
            print("write VTK as points")
            data={"m": (self.mx, self.my, self.mz)}
            if not self.old:
                data.update({"rho":self.pho**self.momentq})
            data.update(kwargs)
            pointsToVTK(vtkname, self.ox, self.oy, self.oz, data=data)#.update(kwargs))
        else : # save as surfaces
            assert len(dim)==3
            print("write VTK as closed surface")
            if close:
                # manually close the gap
                phi = 2*np.pi/self.nfp
                def map_toroidal(vec):
                    rotate = np.array([[  np.cos(phi), np.sin(phi), 0], \
                                       [ -np.sin(phi), np.cos(phi), 0], \
                                       [            0,           0, 1]])
                    return np.matmul(vec, rotate)                
                data_array = {"ox":self.ox, "oy":self.oy, "oz":self.oz, \
                              "mx":self.mx, "my":self.my, "mz": self.mz, \
                              "Ic":self.Ic, "rho":self.pho**self.momentq}
                data_array.update(kwargs)        
                nr, nz, nt = dim
                for key in list(data_array.keys()):
                    new_vec = np.zeros((nr, nz+1, nt+1))
                    for ir in range(nr):
                        new_vec[ir,:,:] = map_matrix(np.reshape(data_array[key], dim)[ir,:,:])
                    if toroidal :
                        data_array[key] = new_vec
                    else :
                        if ntnz :
                            data_array[key] = np.ascontiguousarray(new_vec[:,:,:-1])
                        else :
                            data_array[key] = np.ascontiguousarray(new_vec[:,:-1,:])
                ox = np.copy(data_array['ox'])
                oy = np.copy(data_array['oy'])
                oz = np.copy(data_array['oz'])   
                del data_array['ox']
                del data_array['oy']
                del data_array['oz'] 
                data_array['m'] = (data_array['mx'], data_array['my'], data_array['mz'])
                if toroidal and self.nfp>=1 : # not quite sure if should include nfp=1
                    for ir in range(nr):
                        if ntnz:
                            xyz = map_toroidal(np.transpose([ox[ir,:,0], oy[ir,:,0], oz[ir,:,0]]))
                            ox[ir,:,nz] = xyz[:,0]
                            oy[ir,:,nz] = xyz[:,1]
                            oz[ir,:,nz] = xyz[:,2]
                            moment = map_toroidal(np.transpose([data_array['mx'][ir,:,0],
                                                                data_array['my'][ir,:,0], 
                                                                data_array['mz'][ir,:,0]]))
                            data_array['m'][0][ir,:,nz] = moment[:,0]
                            data_array['m'][1][ir,:,nz] = moment[:,1]
                            data_array['m'][2][ir,:,nz] = moment[:,2]
                        else : 
                            xyz = map_toroidal(np.transpose([ox[ir,0,:], oy[ir,0,:], oz[ir,0,:]]))
                            ox[ir,nz,:] = xyz[:,0]
                            oy[ir,nz,:] = xyz[:,1]
                            oz[ir,nz,:] = xyz[:,2]
                            moment = map_toroidal(np.transpose([data_array['mx'][ir,0,:],
                                                                data_array['my'][ir,0,:], 
                                                                data_array['mz'][ir,0,:]]))
                            data_array['m'][0][ir,nz,:] = moment[:,0]
                            data_array['m'][1][ir,nz,:] = moment[:,1]
                            data_array['m'][2][ir,nz,:] = moment[:,2]
                del data_array['mx']
                del data_array['my']
                del data_array['mz']
                gridToVTK(vtkname, ox, oy, oz, pointData=data_array)
                return
            else:
                ox = np.reshape(self.ox[:self.num], dim)
                oy = np.reshape(self.oy[:self.num], dim)
                oz = np.reshape(self.oz[:self.num], dim)
                mx = np.reshape(self.mx[:self.num], dim)
                my = np.reshape(self.my[:self.num], dim)
                mz = np.reshape(self.mz[:self.num], dim)
                rho = np.reshape(self.pho[:self.num]**self.momentq, dim)
                Ic = np.reshape(self.Ic[:self.num], dim)
            data = {"m": (mx, my, mz), "rho":rho, "Ic":Ic}
            data.update(kwargs)
            gridToVTK(vtkname, ox, oy, oz, pointData=data)
        return

    def full_period(self, nfp=1, dim=None):
        """
        map from one period to full periods
        """
        assert nfp>=1
        self.nfp = nfp
        if not self.xyz_switch:
            self.sp2xyz()
        # change order
        if dim is not None :
            self.ox = np.ravel(np.transpose(np.reshape(self.ox, dim)[::-1,::-1,::-1], (2,0,1)))
            self.oy = np.ravel(np.transpose(np.reshape(self.oy, dim)[::-1,::-1,::-1], (2,0,1)))
            self.oz = np.ravel(np.transpose(np.reshape(self.oz, dim)[::-1,::-1,::-1], (2,0,1)))
            self.mx = np.ravel(np.transpose(np.reshape(self.mx, dim)[::-1,::-1,::-1], (2,0,1)))
            self.my = np.ravel(np.transpose(np.reshape(self.my, dim)[::-1,::-1,::-1], (2,0,1)))
            self.mz = np.ravel(np.transpose(np.reshape(self.mz, dim)[::-1,::-1,::-1], (2,0,1)))
            self.mm = np.ravel(np.transpose(np.reshape(self.mm, dim)[::-1,::-1,::-1], (2,0,1)))
            self.Ic = np.ravel(np.transpose(np.reshape(self.Ic, dim)[::-1,::-1,::-1], (2,0,1)))
            self.Lc = np.ravel(np.transpose(np.reshape(self.Lc, dim)[::-1,::-1,::-1], (2,0,1)))
            self.pho= np.ravel(np.transpose(np.reshape(self.pho,dim)[::-1,::-1,::-1], (2,0,1)))
        if np.mean(self.symm) == 2:
            # get the stellarator symmetry part first
            # Here, we assume no dipoles on the symmetry plane, or only half are listed.
            self.num *= 2
            self.name = np.concatenate((self.name , self.name[::-1]))
            self.ox = np.concatenate((self.ox , self.ox[::-1]))
            self.oy = np.concatenate((self.oy , self.oy[::-1]*(-1)))
            self.oz = np.concatenate((self.oz , self.oz[::-1]*(-1)))
            self.mx = np.concatenate((self.mx , self.mx[::-1]*(-1)))
            self.my = np.concatenate((self.my , self.my[::-1]))
            self.mz = np.concatenate((self.mz , self.mz[::-1]))
            self.mm = np.concatenate((self.mm , self.mm[::-1]))
            self.pho= np.concatenate((self.pho, self.pho[::-1]))
            self.Ic = np.concatenate((self.Ic , self.Ic[::-1]))
            self.Lc = np.concatenate((self.Lc , self.Lc[::-1]))
        xyz = toroidal_period(np.transpose([self.ox, self.oy, self.oz]), self.nfp)
        self.ox = xyz[:,0].copy()
        self.oy = xyz[:,1].copy()
        self.oz = xyz[:,2].copy()
        moment = toroidal_period(np.transpose([self.mx, self.my, self.mz]), self.nfp)
        self.mx = moment[:,0].copy()
        self.my = moment[:,1].copy()
        self.mz = moment[:,2].copy()
        self.mm = np.tile(self.mm, self.nfp)
        self.pho = np.tile(self.pho, self.nfp)
        self.Ic = np.tile(self.Ic, self.nfp)
        self.Lc = np.tile(self.Lc, self.nfp)
        self.name = np.tile(self.name, self.nfp)
        self.num *= self.nfp
        self.symm = np.zeros(self.num, dtype=int)
        self.sp_switch = False  
        return
    
    def inverse(self):
        """get the stellarator symmetric part?
        """        
        if not self.xyz_switch:
            self.sp2xyz()
        self.ox = np.copy(self.ox[::-1])
        self.oy = np.copy(self.oy[::-1]*(-1))
        self.oz = np.copy(self.oz[::-1]*(-1))
        self.mx = np.copy(self.mx[::-1]*(-1))
        self.my = np.copy(self.my[::-1])
        self.mz = np.copy(self.mz[::-1])
        self.mm = np.copy(self.mm[::-1])
        self.pho= np.copy(self.pho[::-1])
        self.Ic = np.copy(self.Ic[::-1])
        self.Lc = np.copy(self.Lc[::-1])
        return
        
    def change_momentq(self, newq):
        """
        change the q factor for the normalized density
        """
        assert newq>0
        pho = self.pho**self.momentq
        # get signs
        self.sp2xyz()
        sign = np.sign(pho)
        if newq%2 == 0: # even number, flip the orientation when negative
            #self.mx *= sign
            #self.my *= sign
            #self.mz *= sign
            self.momentq = newq
            self.xyz2sp()
            # convert to positive rho
            #self.pho = np.power(np.abs(pho), 1.0/newq)
        else: # odd exponetial index
            self.momentq = newq
            # convert to positive rho
            self.pho = np.power(np.abs(pho), 1.0/newq)
            self.pho *= sign
        return

    def plot_rho_profile(self, lower=0, upper=1, nrange=10, nofigure=False, **kwargs):
        import matplotlib.pyplot as plt
        pho = np.abs(self.pho**self.momentq)
        zone = np.linspace(lower, upper, nrange+1, endpoint=True)
        count = []
        for i in range(nrange-1):
            count.append(((pho>=zone[i]) & (pho<zone[i+1])).sum())
        count.append(((pho>=zone[nrange-1]) & (pho<=zone[nrange])).sum())
        count = np.array(count)
        if not nofigure:
            if plt.get_fignums():
                fig = plt.gcf()
                ax = plt.gca()
            else :
                fig, ax = plt.subplots()
            plt.bar(zone[:-1], 100*count/float(self.num), width=0.9/nrange, **kwargs)
            ax.set_xlabel(r'$|\rho|$', fontsize=15)
            ax.set_ylabel('fraction [%]', fontsize=15)
            plt.xticks(fontsize=14)
            plt.yticks(fontsize=14)
        return count

    def volume(self, magnitization=1.1E6, **kwargs):
        self.total_moment = np.sum(np.abs(self.rho*self.mm))
        return self.total_moment/magnitization

    def orientation(self, unit=True, uniform=False):
        oldq = self.momentq
        self.change_momentq(2) # to a even number
        self.change_momentq(oldq) # recover
        if unit: 
            self.mm[:] = 1.
        if uniform:
            self.pho[:] = 1.
        self.sp2xyz()
        return

    def plot(self, start=0, end=None, **kwargs):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        if end is None:
            end = self.num
        fig = plt.figure() 
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(self.ox[start:end], self.oy[start:end], self.oz[start:end], **kwargs)
        return ax

    def __repr__(self):
        return 'FAMUS dipole class, num={:d}, symmetry={:}, filename={:}'.format(
                self.num, np.mean(self.symm), self.filename)

    def __add__(self, other):
        """ Combine two dipole files together
        
        """
        if not self.sp_switch:
            self.xyz2sp()
        assert self.momentq == other.momentq, "Two classes should use the same momentq."
        return Dipole(symm=np.concatenate((self.symm, other.symm)),
                      name=np.concatenate((self.name, other.name)),
                      ox=np.concatenate((self.ox, other.ox)),
                      oy=np.concatenate((self.oy, other.oy)),
                      oz=np.concatenate((self.oz, other.oz)),
                      Ic=np.concatenate((self.Ic, other.Ic)),
                      mm=np.concatenate((self.mm, other.mm)),
                      Lc=np.concatenate((self.Lc, other.Lc)),
                      mp=np.concatenate((self.mp, other.mp)),
                      mt=np.concatenate((self.mt, other.mt)),
                      pho=np.concatenate((self.pho, other.pho)),
                      momentq=self.momentq, 
                      filename=self.filename + '+' + other.filename)
        
    def __del__(self):
        class_name = self.__class__.__name__

class GAdipole(Dipole):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

if __name__ == 'main':
    pass


import setuptools
from coilpy import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='coilpy',
      version=__version__,
      description='Plotting and data processing tools for plasma and coil',
      long_description=long_description,
      long_description_content_type="text/markdown",     
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
      ],
      url='https://github.com/zhucaoxiang/CoilPy',
      author='Caoxiang Zhu',
      author_email='caoxiangzhu@gmail.com',
      license='GNU 3.0',
      packages=setuptools.find_packages(),
)



# modified from stellopt.pySTEL

import numpy as np 
import xarray
import matplotlib.pyplot as plt
from .misc import trig2real
from .surface import FourSurf

__all__ = ['VMECout']

class VMECout(object):
    """
    VMEC wout file

    :param filename: filename passed to OMFITnc class

    All additional key word arguments passed to OMFITnc

    """

    def __init__(self, filename, **kwargs):
        self.wout = xarray.open_dataset(filename, 'r')
        self.data = {}
        self.data['ns'] = int(self.wout['ns'].values)
        self.data['nfp'] = int(self.wout['nfp'].values)
        self.data['nu'] = int(self.wout['mpol'].values*4)
        self.data['nv'] = int(self.wout['ntor'].values*4*self.wout['nfp'].values)
        self.data['nv2'] = int(self.wout['ntor'].values*4)
        self.data['nflux'] = np.linspace(0,1,self.data['ns']) # np.ndarray((self.data['ns'],1))
        self.data['theta'] = np.linspace(0,2*np.pi,self.data['nu']) # np.ndarray((self.data['nu'],1))
        self.data['zeta'] = np.linspace(0,2*np.pi,self.data['nv']) # np.ndarray((self.data['nv'],1))
        self.data['zeta2']= self.data['zeta'][0:self.data['nv2']+1]
        self.surface = []
        self.data['b'] = []
        for i in range(self.data['ns']):
            self.surface.append(FourSurf(xm=self.wout['xm'].values, xn=self.wout['xn'].values,
                                         rbc=self.wout['rmnc'][i].values, rbs=np.zeros_like(self.wout['rmnc'][i].values),
                                         zbs=self.wout['zmns'][i].values, zbc=np.zeros_like(self.wout['zmns'][i].values)))
            self.data['b'].append(trig2real(self.data['theta'], self.data['zeta'],
                                            self.wout['xm_nyq'].values, self.wout['xn_nyq'].values/self.data['nfp'],
                                            self.wout['bmnc'][i].values))
        return

    def plot(self, plot_name='none', ax=None):
        """Plot various VMEC quantities

        Args:
            plot_name (str, optional): The quantity to be plotted, should be one of
                                       iota, q, pressue, <Buco>, <Bvco>, <jcuru>, <jcurv>,
                                       <j.B>, LPK, none. Defaults to 'none'.
            ax (Matplotlib axis, optional): The Matplotlib axis to be plotted on. Defaults to None.
        """        
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()
        if (plot_name == 'none'):
                print("You can plot: iota, q, pressue, <Buco>, <Bvco>, <jcuru>, <jcurv>, ")
                Print("               <j.B>, LPK")
        elif (plot_name == 'iota'):
                ax.plot(self.data['nflux'],self.wout['iotaf'].values)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('iota')
                ax.set_title('Rotational Transform')
                #ax.set(xlabel='s',ylabel='iota',aspect='square')
        elif (plot_name == 'q'):
                ax.plot(self.data['nflux'],1.0/self.wout['iotaf'].values)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('q')
                ax.set_title('Safety Factor')
        elif (plot_name == 'pressure'):
                ax.plot(self.data['nflux'],self.wout['presf'].values/1000)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('Pressure [kPa]')
                ax.set_title('Pressure Profile')
        elif (plot_name == '<Buco>'):
                ax.plot(self.data['nflux'],self.wout['buco'].values)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('<B^u> [T]')
                ax.set_title('Flux surface Averaged B^u')
        elif (plot_name == '<Bvco>'):
                ax.plot(self.data['nflux'],self.wout['bvco'].values)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('<B^v> [T]')
                ax.set_title('Flux surface Averaged B^v')
        elif (plot_name == '<jcuru>'):
                ax.plot(self.data['nflux'],self.wout['jcuru'].values/1000)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('<j^u> [kA/m^2]')
                ax.set_title('Flux surface Averaged j^u')
        elif (plot_name == '<jcurv>'):
                ax.plot(self.data['nflux'],self.wout['jcurv'].values/1000)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('<j^v> [kA/m^2]')
                ax.set_title('Flux surface Averaged j^v')
        elif (plot_name == '<j.B>'):
                ax.plot(self.data['nflux'],self.wout['jdotb'].values/1000)
                ax.set_xlabel('Normalized Flux')
                ax.set_ylabel('<j.B> [T*kA/m^2]')
                ax.set_title('Flux surface Averaged j.B')
        elif (plot_name == 'LPK'):
                self.surface[-1].plot(zeta=0, color='red', label=r'$\phi=0$')
                self.surface[-1].plot(zeta=0.5*np.pi/self.data['nfp'], color='green', label=r'$\phi=0.25$')
                self.surface[-1].plot(zeta=np.pi/self.data['nfp'], color='blue', label=r'$\phi=0.5$')
                ax.set_title('LPK Plot')
        elif (plot_name[0] == '-'):
                print(plot_name)
        else:
            return



import numpy as np

class FourSurf(object):
    '''
    toroidal surface in Fourier representation
    R = \sum RBC cos(mu-nv) + RBS sin(mu-nv)
    Z = \sum ZBC cos(mu-nv) + ZBS sin(mu-nv)
    '''
    def __init__(self, xm=[], xn=[], rbc=[], zbs=[], rbs=[], zbc=[]):
        """Initialization with Fourier harmonics.
        
        Parameters:
          xm -- list or numpy array, array of m index (default: [])
          xn -- list or numpy array, array of n index (default: [])
          rbc -- list or numpy array, array of radial cosine harmonics (default: [])
          zbs -- list or numpy array, array of z sine harmonics (default: [])   
          rbs -- list or numpy array, array of radial sine harmonics (default: [])
          zbc -- list or numpy array, array of z cosine harmonics (default: [])      
        
        """
        self.xm  = np.atleast_1d(xm)
        self.xn  = np.atleast_1d(xn)
        self.rbc = np.atleast_1d(rbc)
        self.rbs = np.atleast_1d(rbs)
        self.zbc = np.atleast_1d(zbc)
        self.zbs = np.atleast_1d(zbs)
        self.mn = len(self.xn)
        return

    @classmethod
    def read_focus_input(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the FOCUS format input file 'plasma.boundary'
        
        Parameters:
          filename -- string, path + name to the FOCUS input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)
        
        Returns:
          fourier_surface class
        """
        with open(filename, 'r') as f:
            line = f.readline() #skip one line
            line = f.readline()
            num = int(line.split()[0]) #harmonics number
            nfp = int(line.split()[1]) #number of field periodicity
            nbn = int(line.split()[2]) #number of Bn harmonics
            xm = []
            xn = []
            rbc = []
            rbs = []
            zbc = []
            zbs = []
            line = f.readline() #skip one line
            line = f.readline() #skip one line
            for i in range(num):
                line = f.readline()
                line_list = line.split()
                n = int(line_list[0])
                m = int(line_list[1])
                if abs(m)>Mpol or abs(n)>Ntor:
                    continue
                xm.append(m)
                xn.append(n)
                rbc.append(float(line_list[2]))
                rbs.append(float(line_list[3]))
                zbc.append(float(line_list[4]))
                zbs.append(float(line_list[5]))
        return cls(xm=np.array(xm), xn=np.array(xn)*nfp,
                       rbc=np.array(rbc), rbs=np.array(rbs),
                       zbc=np.array(zbc), zbs=np.array(zbs))

    @classmethod
    def read_spec_input(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the SPEC input file '*.sp'
        
        Parameters:
          filename -- string, path + name to the FOCUS input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)
        
        Returns:
          fourier_surface class
        """
        import f90nml
        from misc import vmecMN
        spec = f90nml(filename)
        # spec['physicslist'] = 
        Mpol = min(Mpol, spec['physicslist']['MPOL'])
        Ntor = min(Ntor, spec['physicslist']['NTOR'])
        xm, xn = vmecMN(Mpol, Ntor)
        return
        

    @classmethod
    def read_spec_output(cls, spec_out, ns=-1):
        """initialize surface from the ns-th interface SPEC output 
        
        Parameters:
          spec_out -- SPEC class, SPEC hdf5 results
          ns -- integer, the index of SPEC interface (default: -1)
        
        Returns:
          fourier_surface class
        """
        # check if spec_out is in correct format
        #if not isinstance(spec_out, SPEC):
        #    raise TypeError("Invalid type of input data, should be SPEC type.")
        # get required data
        xm = spec_out.output.im
        xn = spec_out.output.in1
        rbc = spec_out.output.Rbc[ns,:]
        zbs = spec_out.output.Zbs[ns,:]
        if spec_out.input.physics.Istellsym:
            # stellarator symmetry enforced
            rbs = np.zeros_like(rbc)
            zbc = np.zeros_like(rbc)
        else:
            rbs = spec_out.output.Rbs[ns,:]
            zbc = spec_out.output.Zbc[ns,:]
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_vmec_input(cls, filename, tol=1E-8):
        """initialize surface from the ns-th interface SPEC output 
        
        Parameters:
          woutfile -- string, path + name to the wout file from VMEC output
          ns -- integer, the index of VMEC nested flux surfaces (default: -1)
        
        Returns:
          fourier_surface class
        """
        import f90nml
        nml = f90nml.read(filename)
        indata = nml['indata']
        mpol = indata['mpol']
        ntor = indata['ntor']
        nfp = indata['nfp']
        arr_rbc = np.array(indata['rbc'])
        arr_zbs = np.array(indata['zbs'])
        arr_rbc[arr_rbc==None] = 0
        arr_zbs[arr_zbs==None] = 0
        try:
            arr_rbs = np.array(indata['rbs'])
            arr_zbc = np.array(indata['zbc'])
            arr_rbs[arr_rbs==None] = 0
            arr_zbc[arr_zbc==None] = 0
        except KeyError:
            arr_rbs = np.zeros_like(arr_rbc)
            arr_zbc = np.zeros_like(arr_rbc)
        nmin, mmin = indata.start_index['rbc']
        mlen, nlen = np.shape(indata['rbc'])
        xm = [] ; xn = []
        rbc = [] ; zbs = []
        rbs = [] ; zbc = []
        for i in range(mlen):
            m = i + mmin
            if m > mpol:
                continue
            for j in range(nlen):
                n = j + nmin
                if n > ntor:
                    continue
                if abs(arr_rbc[i,j])+abs(arr_zbs[i,j])\
                  +abs(arr_rbs[i,j])+abs(arr_zbc[i,j])<tol:
                    continue
                xm.append(m)
                xn.append(n*nfp)
                rbc.append(arr_rbc[i,j])
                zbs.append(arr_zbs[i,j])
                rbs.append(arr_rbs[i,j])
                zbc.append(arr_zbc[i,j])
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_vmec_output(cls, woutfile, ns=-1):
        """initialize surface from the ns-th interface SPEC output 
        
        Parameters:
          woutfile -- string, path + name to the wout file from VMEC output
          ns -- integer, the index of VMEC nested flux surfaces (default: -1)
        
        Returns:
          fourier_surface class
        """
        import xarray as ncdata # read netcdf file
        vmec = ncdata.open_dataset(woutfile)
        xm = vmec['xm'].values
        xn = vmec['xn'].values
        rmnc = vmec['rmnc'].values
        zmns = vmec['zmns'].values
        rbc = rmnc[ns,:]
        zbs = zmns[ns,:]

        if vmec['lasym__logical__'].values:
            # stellarator symmetry enforced
            zmnc = vmec['zmnc'].values
            rmns = vmec['rmns'].values
            rbs = rmns[ns,:]
            zbc = zmnc[ns,:]
        else :
            rbs = np.zeros_like(rbc)
            zbc = np.zeros_like(rbc)
        return cls(xm=xm, xn=xn, rbc=rbc, rbs=rbs, zbc=zbc, zbs=zbs)

    @classmethod
    def read_winding_surfce(cls, filename, Mpol=9999, Ntor=9999):
        """initialize surface from the NESCOIL format input file 'nescin.xxx'
        
        Parameters:
          filename -- string, path + name to the NESCOIL input boundary file
          Mpol -- maximum truncated poloidal mode number (default: 9999)
          Ntol -- maximum truncated toroidal mode number (default: 9999)
        
        Returns:
          fourier_surface class
        """
        with open(filename, 'r') as f:
            line = ''
            while "phip_edge" not in line:
                line = f.readline()
            line = f.readline()
            nfp = int(line.split()[0])
            #print "nfp:",nfp

            line = ''
            while "Current Surface" not in line:
                line = f.readline()
            line = f.readline()
            line = f.readline()
            #print "Number of Fourier modes in coil surface from nescin file: ",line
            num = int(line)
            xm = []
            xn = []
            rbc = []
            rbs = []
            zbc = []
            zbs = []
            line = f.readline() #skip one line
            line = f.readline() #skip one line
            for i in range(num):
                line = f.readline()
                line_list = line.split()
                m = int(line_list[0])
                n = int(line_list[1])
                if abs(m)>Mpol or abs(n)>Ntor:
                    continue
                xm.append(m)
                xn.append(n)
                rbc.append(float(line_list[2]))
                zbs.append(float(line_list[3]))
                rbs.append(float(line_list[4]))
                zbc.append(float(line_list[5]))
            # NESCOIL uses mu+nv, minus sign is added
            return cls(xm=np.array(xm), xn=-np.array(xn)*nfp,
                       rbc=np.array(rbc), rbs=np.array(rbs),
                       zbc=np.array(zbc), zbs=np.array(zbs))

    def rz(self, theta, zeta, normal=False):
        """ get r,z position of list of (theta, zeta)
        
        Parameters:
          theta -- float array_like, poloidal angle
          zeta -- float array_like, toroidal angle value
          normal -- logical, calculate the normal vector or not (default: False)

        Returns:
           r, z -- float array_like
           r, z, [rt, zt], [rz, zz] -- if normal
        """
        assert len(np.atleast_1d(theta)) == len(np.atleast_1d(zeta)), "theta, zeta should be equal size"
        # mt - nz (in matrix)
        _mtnz = np.matmul( np.reshape(self.xm, (-1,1)), np.reshape(theta, (1,-1)) ) \
              - np.matmul( np.reshape(self.xn, (-1,1)), np.reshape( zeta, (1,-1)) ) 
        _cos = np.cos(_mtnz)
        _sin = np.sin(_mtnz)

        r = np.matmul( np.reshape(self.rbc, (1,-1)), _cos ) \
          + np.matmul( np.reshape(self.rbs, (1,-1)), _sin )
        z = np.matmul( np.reshape(self.zbc, (1,-1)), _cos ) \
          + np.matmul( np.reshape(self.zbs, (1,-1)), _sin )

        if not normal :
            return (r.ravel(), z.ravel())
        else:
            rt = np.matmul( np.reshape(self.xm * self.rbc, (1,-1)), -_sin ) \
               + np.matmul( np.reshape(self.xm * self.rbs, (1,-1)),  _cos )
            zt = np.matmul( np.reshape(self.xm * self.zbc, (1,-1)), -_sin ) \
               + np.matmul( np.reshape(self.xm * self.zbs, (1,-1)),  _cos )

            rz = np.matmul( np.reshape(-self.xn * self.rbc, (1,-1)), -_sin ) \
               + np.matmul( np.reshape(-self.xn * self.rbs, (1,-1)),  _cos )
            zz = np.matmul( np.reshape(-self.xn * self.zbc, (1,-1)), -_sin ) \
               + np.matmul( np.reshape(-self.xn * self.zbs, (1,-1)),  _cos )
            return (r.ravel(), z.ravel(), [rt.ravel(), zt.ravel()], [rz.ravel(), zz.ravel()])

    def xyz(self, theta, zeta, normal=False):
        """ get x,y,z position of list of (theta, zeta)
        
        Parameters:
          theta -- float array_like, poloidal angle
          zeta -- float array_like, toroidal angle value
          normal -- logical, calculate the normal vector or not (default: False)

        Returns:
           x, y, z -- float array_like
           x, y, z, [nx, ny, nz] -- if normal
        """
        data = self.rz(theta, zeta, normal)
        r = data[0]
        z = data[1]
        _sin = np.sin(np.ravel(zeta))
        _cos = np.cos(np.ravel(zeta))                      
        if not normal:
            return (r*_cos, r*_sin, z)
        else:
            _xt = data[2][0]*_cos # dx/dtheta
            _yt = data[2][0]*_sin # dy/dtheta
            _zt = data[2][1]      # dz/dtheta
            _xz = data[3][0]*_cos - r*_sin # dx/dzeta
            _yz = data[3][0]*_sin + r*_cos # dy/dzeta
            _zz = data[3][1]               # dz/dzeta
            # n = dr/dz x  dr/dt
            n = np.cross(np.transpose([_xz, _yz, _zz]), np.transpose([_xt, _yt, _zt]))
            return (r*_cos, r*_sin, z, n)

    def _areaVolume(self, theta0=0.0, theta1=2*np.pi, zeta0=0.0, zeta1=2*np.pi, \
                   npol=360, ntor=360):
        """ Internel function to get surface area and volume
        
        Parameters:
          theta0 -- float, starting poloidal angle (default: 0.0)
          theta1 -- float, ending poloidal angle (default: 2*np.pi)
          zeta0 -- float, starting toroidal angle (default: 0.0)
          zeta1 -- float, ending toroidal angle (default: 2*np.pi)
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)

        Returns:
          area -- surface area
          volume -- surface volume
        """
        # get mesh data
        _theta = np.linspace(theta0, theta1, npol, endpoint=False)
        _zeta = np.linspace(zeta0, zeta1, ntor, endpoint=False)
        _tv, _zv = np.meshgrid(_theta, _zeta, indexing='ij')
        _x, _y, _z, _n = self.xyz(_tv, _zv, normal=True)
        # calculates the area and volume
        _dt = (theta1-theta0)/npol
        _dz = (zeta1 - zeta0)/ntor
        _nn = np.linalg.norm(_n, axis=1)
        area = np.sum(_nn)*_dt*_dz
        volume = abs(np.sum(_x*_n[:,0]))*_dt*_dz
        return area, volume
        
    def get_area(self):
        """ Get the surface area and saved in self.area
        More comprehensive options can be found in self._areaVolume()

        Parameters:       
           None

        Returns:
           area
        """        
        self.area, _volume = self._areaVolume()
        return self.area
    
    def get_volume(self):
        """ Get the surface volume and saved in self.volume
        More comprehensive options can be found in self._areaVolume()

        Parameters:       
           None

        Returns:
           volume
        """        
        _area, self.volume = self._areaVolume()
        return self.volume
    
    def plot(self, zeta=0.0, npoints=360, **kwargs):
        """ plot the cross-section at zeta using matplotlib.pyplot
        
        Parameters:       
          zeta -- float, toroidal angle value
          npoints -- integer, number of discretization points (default: 360)
          kwargs -- optional keyword arguments for pyplot

        Returns:
           line class in matplotlib.pyplot
        """
        import matplotlib.pyplot as plt
        # get figure and ax data
        if plt.get_fignums():
            fig = plt.gcf()
            ax = plt.gca()
        else :
            fig, ax = plt.subplots()
        # set default plotting parameters
        if kwargs.get('linewidth') == None:
            kwargs.update({'linewidth': 2.0}) # prefer thicker lines
        if kwargs.get('label') == None:
            kwargs.update({'label': 'toroidal surface'}) # default label 
        # get (r,z) data
        _r, _z = self.rz( np.linspace(0, 2*np.pi, npoints), zeta*np.ones(npoints) )
        line = ax.plot(_r, _z, **kwargs)
        plt.axis('equal')
        plt.xlabel('R [m]',fontsize=20)
        plt.ylabel('Z [m]',fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        return line

    def plot3d(self, engine='pyplot', theta0=0.0, theta1=2*np.pi, zeta0=0.0, zeta1=2*np.pi, \
                   npol=360, ntor=360, normal=False, **kwargs):
        """ plot 3D shape of the surface
        
        Parameters: 
          engine -- string, plotting engine {'pyplot' (default), 'mayavi', 'noplot'}
          theta0 -- float, starting poloidal angle (default: 0.0)
          theta1 -- float, ending poloidal angle (default: 2*np.pi)
          zeta0 -- float, starting toroidal angle (default: 0.0)
          zeta1 -- float, ending toroidal angle (default: 2*np.pi)
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)
          kwargs -- optional keyword arguments for plotting

        Returns:
           xsurf, ysurf, zsurf -- arrays of x,y,z coordinates on the surface
        """
        # get mesh data
        _theta = np.linspace(theta0, theta1, npol)
        _zeta = np.linspace(zeta0, zeta1, ntor)
        _tv, _zv = np.meshgrid(_theta, _zeta, indexing='ij')
        if normal:
            _x, _y, _z, _n = self.xyz(_tv, _zv, normal=normal)
            n = np.reshape(_n, (3, npol, ntor))
        else:
            _x, _y, _z = self.xyz(_tv, _zv)
            n = None
        xsurf = np.reshape(_x, (npol, ntor))
        ysurf = np.reshape(_y, (npol, ntor))
        zsurf = np.reshape(_z, (npol, ntor))
        if engine == 'noplot':
            # just return xyz data
            pass
        elif engine == 'pyplot':
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            # plot in matplotlib.pyplot
            if plt.get_fignums():
                fig = plt.gcf()
                ax = plt.gca()
                if ax.name != '3d':
                    ax = fig.add_subplot(111, projection='3d')
            else :
                fig = plt.figure()                
                ax = fig.add_subplot(111, projection='3d')
            ax.plot_surface(xsurf, ysurf, zsurf, **kwargs)
        elif engine == 'mayavi':
            # plot 3D surface in mayavi.mlab
            from mayavi import mlab # to overrid plt.mlab
            mlab.mesh(xsurf, ysurf, zsurf, **kwargs)
        else:
            raise ValueError('Invalid engine option {pyplot, mayavi, noplot}')
        return (xsurf, ysurf, zsurf, n)

    def toVTK(self, vtkname, npol=360, ntor=360, **kwargs):
        """ save surface shape a vtk grid file
        
        Parameters: 
          vtkname -- string, the filename you want to save, final name is 'vtkname.vts'
          npol -- integer, number of poloidal discretization points (default: 360)
          ntor -- integer, number of toroidal discretization points (default: 360)
          kwargs -- optional keyword arguments for saving as pointdata

        Returns:
           
        """
        from pyevtk.hl import gridToVTK # save to binary vtk
        _xx, _yy, _zz, _nn = self.plot3d('noplot', zeta0=0.0, zeta1=2*np.pi,
                                    theta0=0.0, theta1=2*np.pi, npol=npol, ntor=ntor, normal=True)
        _xx = _xx.reshape((1, npol, ntor))
        _yy = _yy.reshape((1, npol, ntor))
        _zz = _zz.reshape((1, npol, ntor))
        _nn = _nn.reshape((1, npol, ntor, 3))

        kwargs.setdefault('n', (np.ascontiguousarray(_nn[:,:,:,0]),
                                np.ascontiguousarray(_nn[:,:,:,1]),
                                np.ascontiguousarray( _nn[:,:,:,2])))
        gridToVTK(vtkname, _xx, _yy, _zz, pointData=kwargs)
        return    

    def write_focus_input(self, filename, Nfp=1, tol=1E-8, **kwargs):
        ''' write the Fourier harmonics down in FOCUS format

        Parameters:
          filename -- string, the saved FOCUS boundary name
          Nfp -- int, default: 1, number of field periodicity, n will be divided by Nfp
          tol -- float, default: 1E-8, tolerance for truncating modes
        '''
        # count non-zero
        nonzero = abs(self.rbc)+abs(self.rbs)+abs(self.zbc)+abs(self.zbs)
        amn = np.sum(nonzero>tol)
        # write Fourier coefficients
        with open(filename, 'w') as fofile:
            fofile.write('# bmn   bNfp   nbf '+'\n')
            fofile.write("{:d} \t {:d} \t {:d} \n".format(amn, Nfp, 0))
            fofile.write('#plasma boundary'+'\n')
            fofile.write('# n m Rbc Rbs Zbc Zbs'+'\n')
            for imn in range(self.mn):
                if nonzero[imn] > tol :
                    fofile.write("{:4d}  {:4d} \t {:23.15E}  {:23.15E}  {:23.15E}  {:23.15E} \n".format(
                            self.xn[imn]//Nfp, self.xm[imn], self.rbc[imn], self.rbs[imn], self.zbc[imn], self.zbs[imn]))
            fofile.write("#Bn harmonics \n")
            fofile.write('# n m bnc bns \n')         
        print("Finished write FOCUS input file at ", filename)
        return        

    def __del__(self):
        class_name = self.__class__.__name__
